import logging
from collections import OrderedDict
from socket import AF_INET, AF_INET6

from pr2modules.common import basestring
from pr2modules.netlink.rtnl.fibmsg import FR_ACT_NAMES
from pr2modules.netlink.rtnl.ifinfmsg import ifinfmsg, protinfo_bridge

log = logging.getLogger(__name__)


class IPRequest(OrderedDict):
    def __init__(self, obj=None, command=None):
        super(IPRequest, self).__init__()
        self.command = command
        if obj is not None:
            self.update(obj)

    def update(self, obj):
        if obj.get('family', None):
            self['family'] = obj['family']
        for key in obj:
            if key == 'family':
                continue
            v = obj[key]
            if isinstance(v, dict):
                self[key] = dict((x for x in v.items() if x[1] is not None))
            elif v is not None:
                self[key] = v
        self.fix_request()

    def fix_request(self):
        pass

    def set(self, key, value):
        return super(IPRequest, self).__setitem__(key, value)

    def sync_cacheinfo(self):
        pass


class IPRuleRequest(IPRequest):
    def fix_request(self):
        # now fix the rest
        if 'family' not in self:
            self['family'] = AF_INET
        if ('priority' not in self) and ('FRA_PRIORITY' not in self):
            self['priority'] = 32000
        if 'table' in self and 'action' not in self:
            self['action'] = 'to_tbl'
        for key in ('src_len', 'dst_len'):
            if self.get(key, None) is None and key[:3] in self:
                self[key] = {AF_INET6: 128, AF_INET: 32}[self['family']]

    def __setitem__(self, key, value):
        if key.startswith('ipdb_'):
            return

        if key in ('src', 'dst'):
            v = value.split('/')
            if len(v) == 2:
                value, self['%s_len' % key] = v[0], int(v[1])
        elif key == 'action' and isinstance(value, basestring):
            value = FR_ACT_NAMES.get(
                value, (FR_ACT_NAMES.get('FR_ACT_' + value.upper(), value))
            )

        self.set(key, value)


class CBRequest(IPRequest):
    '''
    FIXME
    '''

    commands = None
    msg = None

    def __init__(self, *argv, **kwarg):
        self['commands'] = {'attrs': []}

    def __setitem__(self, key, value):
        if value is None:
            return
        if key in self.commands:
            self['commands']['attrs'].append([self.msg.name2nla(key), value])
        else:
            self.set(key, value)


class IPBridgeRequest(IPRequest):
    def __setitem__(self, key, value):
        if key in ('vlan_info', 'mode', 'vlan_flags'):
            if 'IFLA_AF_SPEC' not in self:
                (
                    super(IPBridgeRequest, self).__setitem__(
                        'IFLA_AF_SPEC', {'attrs': []}
                    )
                )
            nla = ifinfmsg.af_spec_bridge.name2nla(key)
            self['IFLA_AF_SPEC']['attrs'].append([nla, value])
        else:
            self.set(key, value)


class IPBrPortRequest(dict):
    def __init__(self, obj=None):
        dict.__init__(self)
        dict.__setitem__(self, 'attrs', [])
        self.allowed = [x[0] for x in protinfo_bridge.nla_map]
        if obj is not None:
            self.update(obj)

    def update(self, obj):
        for key in obj:
            self[key] = obj[key]

    def __setitem__(self, key, value):
        key = protinfo_bridge.name2nla(key)
        if key in self.allowed:
            self['attrs'].append((key, value))



import logging

from pr2modules.netlink import nla
from pr2modules.netlink.rtnl import TC_H_ROOT
from pr2modules.netlink.rtnl.tcmsg.common import get_time
from pr2modules.netlink.rtnl.tcmsg.common import stats2 as c_stats2

log = logging.getLogger(__name__)
parent = TC_H_ROOT


def get_parameters(kwarg):
    #
    # ACHTUNG: experimental code
    #
    # Parameters naming scheme WILL be changed in next releases
    #
    ret = {'attrs': []}
    transform = {
        'cdl_limit': lambda x: x,
        'cdl_ecn': lambda x: x,
        'cdl_target': get_time,
        'cdl_ce_threshold': get_time,
        'cdl_interval': get_time,
    }
    for key in transform:
        if key in kwarg:
            log.warning(
                'codel parameters naming will be changed '
                'in next releases (%s)' % key
            )
            ret['attrs'].append(
                ['TCA_CODEL_%s' % key[4:].upper(), transform[key](kwarg[key])]
            )
    return ret


class options(nla):
    nla_map = (
        ('TCA_CODEL_UNSPEC', 'none'),
        ('TCA_CODEL_TARGET', 'uint32'),
        ('TCA_CODEL_LIMIT', 'uint32'),
        ('TCA_CODEL_INTERVAL', 'uint32'),
        ('TCA_CODEL_ECN', 'uint32'),
        ('TCA_CODEL_CE_THRESHOLD', 'uint32'),
    )


class stats(nla):
    fields = (
        ('maxpacket', 'I'),
        ('count', 'I'),
        ('lastcount', 'I'),
        ('ldelay', 'I'),
        ('drop_next', 'I'),
        ('drop_overlimit', 'I'),
        ('ecn_mark', 'I'),
        ('dropping', 'I'),
        ('ce_mark', 'I'),
    )


class stats2(c_stats2):
    class stats_app(stats):
        pass



from pr2modules.netlink.rtnl.tcmsg.common import (
    get_rate_parameters,
    nla_plus_rtab,
)

actions = {
    'unspec': -1,  # TC_POLICE_UNSPEC
    'ok': 0,  # TC_POLICE_OK
    'reclassify': 1,  # TC_POLICE_RECLASSIFY
    'shot': 2,  # TC_POLICE_SHOT
    'drop': 2,  # TC_POLICE_SHOT
    'pipe': 3,
}  # TC_POLICE_PIPE


class options(nla_plus_rtab):
    nla_map = (
        ('TCA_POLICE_UNSPEC', 'none'),
        ('TCA_POLICE_TBF', 'police_tbf'),
        ('TCA_POLICE_RATE', 'rtab'),
        ('TCA_POLICE_PEAKRATE', 'ptab'),
        ('TCA_POLICE_AVRATE', 'uint32'),
        ('TCA_POLICE_RESULT', 'uint32'),
    )

    class police_tbf(nla_plus_rtab.parms):
        fields = (
            ('index', 'I'),
            ('action', 'i'),
            ('limit', 'I'),
            ('burst', 'I'),
            ('mtu', 'I'),
            ('rate_cell_log', 'B'),
            ('rate___reserved', 'B'),
            ('rate_overhead', 'H'),
            ('rate_cell_align', 'h'),
            ('rate_mpu', 'H'),
            ('rate', 'I'),
            ('peak_cell_log', 'B'),
            ('peak___reserved', 'B'),
            ('peak_overhead', 'H'),
            ('peak_cell_align', 'h'),
            ('peak_mpu', 'H'),
            ('peak', 'I'),
            ('refcnt', 'i'),
            ('bindcnt', 'i'),
            ('capab', 'I'),
        )


class nla_plus_police(object):
    class police(options):
        pass


def get_parameters(kwarg):
    # if no limit specified, set it to zero to make
    # the next call happy
    kwarg['limit'] = kwarg.get('limit', 0)
    tbfp = get_rate_parameters(kwarg)
    # create an alias -- while TBF uses 'buffer', rate
    # policy uses 'burst'
    tbfp['burst'] = tbfp['buffer']
    # action resolver
    tbfp['action'] = actions[kwarg.get('action', 'reclassify')]
    return {'attrs': [['TCA_POLICE_TBF', tbfp], ['TCA_POLICE_RATE', True]]}



from pr2modules.netlink import nla
from pr2modules.netlink.rtnl.tcmsg import (
    act_bpf,
    act_connmark,
    act_gact,
    act_mirred,
    act_police,
    act_skbedit,
    act_vlan,
)
from pr2modules.netlink.rtnl.tcmsg.common import TCA_ACT_MAX_PRIO, stats2

plugins = {
    'gact': act_gact,
    'bpf': act_bpf,
    'police': act_police,
    'mirred': act_mirred,
    'connmark': act_connmark,
    'vlan': act_vlan,
    'skbedit': act_skbedit,
}


class nla_plus_tca_act_opt(object):
    @staticmethod
    def get_act_options(self, *argv, **kwarg):
        kind = self.get_attr('TCA_ACT_KIND')
        if kind in plugins:
            return plugins[kind].options

        return self.hex


class tca_act_prio(nla):
    nla_map = tuple(
        [('TCA_ACT_PRIO_%i' % x, 'tca_act') for x in range(TCA_ACT_MAX_PRIO)]
    )

    class tca_act(nla, nla_plus_tca_act_opt):
        nla_map = (
            ('TCA_ACT_UNSPEC', 'none'),
            ('TCA_ACT_KIND', 'asciiz'),
            ('TCA_ACT_OPTIONS', 'get_act_options'),
            ('TCA_ACT_INDEX', 'hex'),
            ('TCA_ACT_STATS', 'stats2'),
        )

        stats2 = stats2


def get_act_parms(kwarg):
    if 'kind' not in kwarg:
        raise Exception('action requires "kind" parameter')

    if kwarg['kind'] in plugins:
        return plugins[kwarg['kind']].get_parameters(kwarg)
    else:
        return []


# All filters can use any act type, this is a generic parser for all
def get_tca_action(kwarg):
    ret = {'attrs': []}

    act = kwarg.get('action', 'drop')

    # convert simple action='..' to kwarg style
    if isinstance(act, str):
        act = {'kind': 'gact', 'action': act}

    # convert single dict action to first entry in a list of actions
    acts = act if isinstance(act, list) else [act]

    for i, act in enumerate(acts, start=1):
        opt = {
            'attrs': [
                ['TCA_ACT_KIND', act['kind']],
                ['TCA_ACT_OPTIONS', get_act_parms(act)],
            ]
        }
        ret['attrs'].append(['TCA_ACT_PRIO_%d' % i, opt])

    return ret



from pr2modules.common import map_namespace
from pr2modules.netlink import nla, nlmsg

# neighbor cache entry flags
NTF_USE = 0x01
NTF_SELF = 0x02
NTF_MASTER = 0x04
NTF_PROXY = 0x08
NTF_EXT_LEARNED = 0x10
NTF_ROUTER = 0x80

# neighbor cache entry states
NUD_INCOMPLETE = 0x01
NUD_REACHABLE = 0x02
NUD_STALE = 0x04
NUD_DELAY = 0x08
NUD_PROBE = 0x10
NUD_FAILED = 0x20

# dummy states
NUD_NOARP = 0x40
NUD_PERMANENT = 0x80
NUD_NONE = 0x00

(NTF_NAMES, NTF_VALUES) = map_namespace('NTF_', globals())
(NUD_NAMES, NUD_VALUES) = map_namespace('NUD_', globals())
flags = dict([(x[0][4:].lower(), x[1]) for x in NTF_NAMES.items()])
states = dict([(x[0][4:].lower(), x[1]) for x in NUD_NAMES.items()])


def states_a2n(s):
    # parse state string
    ss = s.split(',')
    ret = 0
    for state in ss:
        state = state.upper()
        if not state.startswith('NUD_'):
            state = 'NUD_' + state
        ret |= NUD_NAMES[state]
    return ret


class ndmsg(nlmsg):
    '''
    ARP cache update message

    C structure::

        struct ndmsg {
            unsigned char ndm_family;
            int           ndm_ifindex;  /* Interface index */
            __u16         ndm_state;    /* State */
            __u8          ndm_flags;    /* Flags */
            __u8          ndm_type;
        };

    Cache info structure::

        struct nda_cacheinfo {
            __u32         ndm_confirmed;
            __u32         ndm_used;
            __u32         ndm_updated;
            __u32         ndm_refcnt;
        };
    '''

    __slots__ = ()

    prefix = 'NDA_'

    fields = (
        ('family', 'B'),
        ('__pad', '3x'),
        ('ifindex', 'i'),
        ('state', 'H'),
        ('flags', 'B'),
        ('ndm_type', 'B'),
    )

    # Please note, that nla_map creates implicit
    # enumeration. In this case it will be:
    #
    # NDA_UNSPEC = 0
    # NDA_DST = 1
    # NDA_LLADDR = 2
    # NDA_CACHEINFO = 3
    # NDA_PROBES = 4
    # ...
    #
    nla_map = (
        ('NDA_UNSPEC', 'none'),
        ('NDA_DST', 'ipaddr'),
        ('NDA_LLADDR', 'lladdr'),
        ('NDA_CACHEINFO', 'cacheinfo'),
        ('NDA_PROBES', 'uint32'),
        ('NDA_VLAN', 'uint16'),
        ('NDA_PORT', 'be16'),
        ('NDA_VNI', 'uint32'),
        ('NDA_IFINDEX', 'uint32'),
        ('NDA_MASTER', 'uint32'),
    )

    class cacheinfo(nla):

        __slots__ = ()

        fields = (
            ('ndm_confirmed', 'I'),
            ('ndm_used', 'I'),
            ('ndm_updated', 'I'),
            ('ndm_refcnt', 'I'),
        )



from pr2modules.netlink import nla


class bond(nla):
    prefix = 'IFLA_'
    nla_map = (
        ('IFLA_BOND_UNSPEC', 'none'),
        ('IFLA_BOND_MODE', 'uint8'),
        ('IFLA_BOND_ACTIVE_SLAVE', 'uint32'),
        ('IFLA_BOND_MIIMON', 'uint32'),
        ('IFLA_BOND_UPDELAY', 'uint32'),
        ('IFLA_BOND_DOWNDELAY', 'uint32'),
        ('IFLA_BOND_USE_CARRIER', 'uint8'),
        ('IFLA_BOND_ARP_INTERVAL', 'uint32'),
        ('IFLA_BOND_ARP_IP_TARGET', '*ipaddr'),
        ('IFLA_BOND_ARP_VALIDATE', 'uint32'),
        ('IFLA_BOND_ARP_ALL_TARGETS', 'uint32'),
        ('IFLA_BOND_PRIMARY', 'uint32'),
        ('IFLA_BOND_PRIMARY_RESELECT', 'uint8'),
        ('IFLA_BOND_FAIL_OVER_MAC', 'uint8'),
        ('IFLA_BOND_XMIT_HASH_POLICY', 'uint8'),
        ('IFLA_BOND_RESEND_IGMP', 'uint32'),
        ('IFLA_BOND_NUM_PEER_NOTIF', 'uint8'),
        ('IFLA_BOND_ALL_SLAVES_ACTIVE', 'uint8'),
        ('IFLA_BOND_MIN_LINKS', 'uint32'),
        ('IFLA_BOND_LP_INTERVAL', 'uint32'),
        ('IFLA_BOND_PACKETS_PER_SLAVE', 'uint32'),
        ('IFLA_BOND_AD_LACP_RATE', 'uint8'),
        ('IFLA_BOND_AD_SELECT', 'uint8'),
        ('IFLA_BOND_AD_INFO', 'ad_info'),
        ('IFLA_BOND_AD_ACTOR_SYS_PRIO', 'uint16'),
        ('IFLA_BOND_AD_USER_PORT_KEY', 'uint16'),
        ('IFLA_BOND_AD_ACTOR_SYSTEM', 'hex'),
        ('IFLA_BOND_TLB_DYNAMIC_LB', 'uint8'),
    )

    class ad_info(nla):
        nla_map = (
            ('IFLA_BOND_AD_INFO_UNSPEC', 'none'),
            ('IFLA_BOND_AD_INFO_AGGREGATOR', 'uint16'),
            ('IFLA_BOND_AD_INFO_NUM_PORTS', 'uint16'),
            ('IFLA_BOND_AD_INFO_ACTOR_KEY', 'uint16'),
            ('IFLA_BOND_AD_INFO_PARTNER_KEY', 'uint16'),
            ('IFLA_BOND_AD_INFO_PARTNER_MAC', 'l2addr'),
        )



import types

from pr2modules.netlink import nla, nlmsg
from pr2modules.netlink.rtnl.tcmsg import (
    cls_basic,
    cls_flow,
    cls_fw,
    cls_matchall,
    cls_u32,
    sched_bpf,
    sched_cake,
    sched_choke,
    sched_clsact,
    sched_codel,
    sched_drr,
    sched_fq_codel,
    sched_hfsc,
    sched_htb,
    sched_ingress,
    sched_netem,
    sched_pfifo,
    sched_pfifo_fast,
    sched_plug,
    sched_sfq,
    sched_tbf,
    sched_template,
)

plugins = {
    'plug': sched_plug,
    'sfq': sched_sfq,
    'clsact': sched_clsact,
    'codel': sched_codel,
    'fq_codel': sched_fq_codel,
    'hfsc': sched_hfsc,
    'htb': sched_htb,
    'bpf': sched_bpf,
    'tbf': sched_tbf,
    'netem': sched_netem,
    'fw': cls_fw,
    'u32': cls_u32,
    'matchall': cls_matchall,
    'basic': cls_basic,
    'flow': cls_flow,
    'ingress': sched_ingress,
    'pfifo': sched_pfifo,
    'pfifo_fast': sched_pfifo_fast,
    'choke': sched_choke,
    'drr': sched_drr,
    'prio': sched_pfifo_fast,
    'cake': sched_cake,
}


class tcmsg(nlmsg):

    prefix = 'TCA_'

    fields = (
        ('family', 'B'),
        ('pad1', 'B'),
        ('pad2', 'H'),
        ('index', 'i'),
        ('handle', 'I'),
        ('parent', 'I'),
        ('info', 'I'),
    )

    nla_map = (
        ('TCA_UNSPEC', 'none'),
        ('TCA_KIND', 'asciiz'),
        ('TCA_OPTIONS', 'get_options'),
        ('TCA_STATS', 'stats'),
        ('TCA_XSTATS', 'get_xstats'),
        ('TCA_RATE', 'hex'),
        ('TCA_FCNT', 'hex'),
        ('TCA_STATS2', 'get_stats2'),
        ('TCA_STAB', 'hex'),
    )

    class stats(nla):
        fields = (
            ('bytes', 'Q'),
            ('packets', 'I'),
            ('drop', 'I'),
            ('overlimits', 'I'),
            ('bps', 'I'),
            ('pps', 'I'),
            ('qlen', 'I'),
            ('backlog', 'I'),
        )

    def get_plugin(self, plug, *argv, **kwarg):
        # get the plugin name
        kind = self.get_attr('TCA_KIND')
        # get the plugin implementation or the default one
        p = plugins.get(kind, sched_template)
        # get the interface
        interface = getattr(p, plug, getattr(sched_template, plug))
        # if it is a method, run and return the result
        if isinstance(interface, types.FunctionType):
            return interface(self, *argv, **kwarg)
        else:
            return interface

    @staticmethod
    def get_stats2(self, *argv, **kwarg):
        return self.get_plugin('stats2', *argv, **kwarg)

    @staticmethod
    def get_xstats(self, *argv, **kwarg):
        return self.get_plugin('stats', *argv, **kwarg)

    @staticmethod
    def get_options(self, *argv, **kwarg):
        return self.get_plugin('options', *argv, **kwarg)



import json
import os
import subprocess

from pr2modules.common import map_enoent
from pr2modules.netlink.rtnl.ifinfmsg import RTM_NEWLINK
from pr2modules.netlink.rtnl.ifinfmsg.sync import sync
from pr2modules.netlink.rtnl.ifinfmsg.tuntap import manage_tun, manage_tuntap

_BONDING_MASTERS = '/sys/class/net/bonding_masters'
_BONDING_SLAVES = '/sys/class/net/%s/bonding/slaves'
_BRIDGE_MASTER = '/sys/class/net/%s/brport/bridge/ifindex'
_BONDING_MASTER = '/sys/class/net/%s/master/ifindex'
IFNAMSIZ = 16


def proxy_setlink(msg, nl):
    def get_interface(index):
        msg = nl.get_links(index)[0]
        try:
            kind = msg.get_attr('IFLA_LINKINFO').get_attr('IFLA_INFO_KIND')
        except AttributeError:
            kind = 'unknown'
        return {
            'ifname': msg.get_attr('IFLA_IFNAME'),
            'master': msg.get_attr('IFLA_MASTER'),
            'kind': kind,
        }

    forward = True

    # is it a port setup?
    master = msg.get_attr('IFLA_MASTER')
    if master is not None:

        if master == 0:
            # port delete
            # 1. get the current master
            iface = get_interface(msg['index'])
            master = get_interface(iface['master'])
            cmd = 'del'
        else:
            # port add
            # 1. get the master
            master = get_interface(master)
            cmd = 'add'

        ifname = (
            msg.get_attr('IFLA_IFNAME')
            or get_interface(msg['index'])['ifname']
        )

        # 2. manage the port
        forward_map = {'team': manage_team_port}
        if master['kind'] in forward_map:
            func = forward_map[master['kind']]
            forward = func(cmd, master['ifname'], ifname, nl)

    if forward is not None:
        return {'verdict': 'forward', 'data': msg.data}


def proxy_newlink(msg, nl):
    kind = None

    # get the interface kind
    linkinfo = msg.get_attr('IFLA_LINKINFO')
    if linkinfo is not None:
        kind = [x[1] for x in linkinfo['attrs'] if x[0] == 'IFLA_INFO_KIND']
        if kind:
            kind = kind[0]

    if kind == 'tuntap':
        return manage_tuntap(msg)
    elif kind == 'tun':
        return manage_tun(msg)
    elif kind == 'team':
        return manage_team(msg)

    return {'verdict': 'forward', 'data': msg.data}


@map_enoent
@sync
def manage_team(msg):

    if msg['header']['type'] != RTM_NEWLINK:
        raise ValueError('wrong command type')

    try:
        linkinfo = msg.get_attr('IFLA_LINKINFO')
        infodata = linkinfo.get_attr('IFLA_INFO_DATA')
        config = infodata.get_attr('IFLA_TEAM_CONFIG')
        config = json.loads(config)
    except AttributeError:
        config = {
            'runner': {'name': 'activebackup'},
            'link_watch': {'name': 'ethtool'},
        }

    # fix device
    config['device'] = msg.get_attr('IFLA_IFNAME')

    with open(os.devnull, 'w') as fnull:
        subprocess.check_call(
            ['teamd', '-d', '-n', '-c', json.dumps(config)],
            stdout=fnull,
            stderr=fnull,
        )


@map_enoent
def manage_team_port(cmd, master, ifname, nl):
    with open(os.devnull, 'w') as fnull:
        subprocess.check_call(
            [
                'teamdctl',
                master,
                'port',
                'remove' if cmd == 'del' else 'add',
                ifname,
            ],
            stdout=fnull,
            stderr=fnull,
        )



from pr2modules.netlink import nla, nlmsg_atoms


class ipoib(nla):
    prefix = 'IFLA_IPOIB_'

    nla_map = (
        ('IFLA_IPOIB_UNSPEC', 'none'),
        ('IFLA_IPOIB_PKEY', 'uint16'),
        ('IFLA_IPOIB_MODE', 'mode'),
        ('IFLA_IPOIB_UMCAST', 'uint16'),
    )

    class mode(nlmsg_atoms.uint16):
        value_map = {0: 'datagram', 1: 'connected'}



from pr2modules.netlink import nla
from pr2modules.netlink.rtnl.tcmsg.common import tc_actions


class options(nla):
    nla_map = (
        ('TCA_ACT_BPF_UNSPEC', 'none'),
        ('TCA_ACT_BPF_TM,', 'none'),
        ('TCA_ACT_BPF_PARMS', 'tca_act_bpf_parms'),
        ('TCA_ACT_BPF_OPS_LEN', 'uint16'),
        ('TCA_ACT_BPF_OPS', 'hex'),
        ('TCA_ACT_BPF_FD', 'uint32'),
        ('TCA_ACT_BPF_NAME', 'asciiz'),
    )

    class tca_act_bpf_parms(nla):
        fields = (
            ('index', 'I'),
            ('capab', 'I'),
            ('action', 'i'),
            ('refcnt', 'i'),
            ('bindcnt', 'i'),
        )


def get_parameters(kwarg):
    ret = {'attrs': []}
    if 'fd' in kwarg:
        ret['attrs'].append(['TCA_ACT_BPF_FD', kwarg['fd']])
    if 'name' in kwarg:
        ret['attrs'].append(['TCA_ACT_BPF_NAME', kwarg['name']])
    a = tc_actions[kwarg.get('action', 'drop')]
    ret['attrs'].append(['TCA_ACT_BPF_PARMS', {'action': a}])
    return ret



'''
Nfnetlink
=========

The support of nfnetlink families is now at the
very beginning. So there is no public exports
yet, but you can review the code. Work is in
progress, stay tuned.

nf-queue
++++++++

Netfilter protocol for NFQUEUE iptables target.
'''

from pr2modules.netlink import nlmsg

NFNL_SUBSYS_NONE = 0
NFNL_SUBSYS_CTNETLINK = 1
NFNL_SUBSYS_CTNETLINK_EXP = 2
NFNL_SUBSYS_QUEUE = 3
NFNL_SUBSYS_ULOG = 4
NFNL_SUBSYS_OSF = 5
NFNL_SUBSYS_IPSET = 6
NFNL_SUBSYS_ACCT = 7
NFNL_SUBSYS_CTNETLINK_TIMEOUT = 8
NFNL_SUBSYS_CTHELPER = 9
NFNL_SUBSYS_NFTABLES = 10
NFNL_SUBSYS_NFT_COMPAT = 11
NFNL_SUBSYS_COUNT = 12

# multicast group ids (for use with {add,drop}_membership)
NFNLGRP_NONE = 0
NFNLGRP_CONNTRACK_NEW = 1
NFNLGRP_CONNTRACK_UPDATE = 2
NFNLGRP_CONNTRACK_DESTROY = 3
NFNLGRP_CONNTRACK_EXP_NEW = 4
NFNLGRP_CONNTRACK_EXP_UPDATE = 5
NFNLGRP_CONNTRACK_EXP_DESTROY = 6
NFNLGRP_NFTABLES = 7
NFNLGRP_ACCT_QUOTA = 8
NFNLGRP_NFTRACE = 9


class nfgen_msg(nlmsg):
    fields = (('nfgen_family', 'B'), ('version', 'B'), ('res_id', '!H'))



'''
TODO: add THERMAL_GENL_ATTR_EVENT structure
'''
from pr2modules.netlink import genlmsg
from pr2modules.netlink.event import EventSocket
from pr2modules.netlink.nlsocket import Marshal

THERMAL_GENL_CMD_UNSPEC = 0
THERMAL_GENL_CMD_EVENT = 1


class thermal_msg(genlmsg):
    nla_map = (
        ('THERMAL_GENL_ATTR_UNSPEC', 'none'),
        ('THERMAL_GENL_ATTR_EVENT', 'hex'),
    )  # to be done


class MarshalThermalEvent(Marshal):
    msg_map = {
        THERMAL_GENL_CMD_UNSPEC: thermal_msg,
        THERMAL_GENL_CMD_EVENT: thermal_msg,
    }


class ThermalEventSocket(EventSocket):
    marshal_class = MarshalThermalEvent
    genl_family = 'thermal_event'



'''
'''
from pr2modules.netlink import genlmsg, nla
from pr2modules.netlink.event import EventSocket
from pr2modules.netlink.nlsocket import Marshal

ACPI_GENL_CMD_UNSPEC = 0
ACPI_GENL_CMD_EVENT = 1


class acpimsg(genlmsg):
    nla_map = (
        ('ACPI_GENL_ATTR_UNSPEC', 'none'),
        ('ACPI_GENL_ATTR_EVENT', 'acpiev'),
    )

    class acpiev(nla):
        fields = (
            ('device_class', '20s'),
            ('bus_id', '15s'),
            ('type', 'I'),
            ('data', 'I'),
        )

        def decode(self):
            nla.decode(self)
            dc = self['device_class']
            bi = self['bus_id']
            self['device_class'] = dc[: dc.find(b'\x00')]
            self['bus_id'] = bi[: bi.find(b'\x00')]


class MarshalAcpiEvent(Marshal):
    msg_map = {ACPI_GENL_CMD_UNSPEC: acpimsg, ACPI_GENL_CMD_EVENT: acpimsg}


class AcpiEventSocket(EventSocket):
    marshal_class = MarshalAcpiEvent
    genl_family = 'acpi_event'



from pr2modules.netlink.rtnl.rtgenmsg import rtgenmsg


class nsidmsg(rtgenmsg):

    nla_map = (
        ('NETNSA_NONE', 'none'),
        ('NETNSA_NSID', 'uint32'),
        ('NETNSA_PID', 'uint32'),
        ('NETNSA_FD', 'uint32'),
        ('NETNSA_TARGET_NSID', 'uint32'),
        ('NETNSA_CURRENT_NSID', 'uint32'),
    )



import logging

from pr2modules.netlink import NLM_F_DUMP, NLM_F_REQUEST
from pr2modules.netlink.devlink import DEVLINK_NAMES, DevlinkSocket, devlinkcmd

log = logging.getLogger(__name__)


class DL(DevlinkSocket):
    def __init__(self, *argv, **kwarg):
        # get specific groups kwarg
        if 'groups' in kwarg:
            groups = kwarg['groups']
            del kwarg['groups']
        else:
            groups = None

        # get specific async kwarg
        if 'async' in kwarg:
            # FIXME
            # raise deprecation error after 0.5.3
            #
            log.warning(
                'use "async_cache" instead of "async", '
                '"async" is a keyword from Python 3.7'
            )
            kwarg['async_cache'] = kwarg.pop('async')

        if 'async_cache' in kwarg:
            async_cache = kwarg.pop('async_cache')
        else:
            async_cache = False

        # align groups with async_cache
        if groups is None:
            groups = ~0 if async_cache else 0

        # continue with init
        super(DL, self).__init__(*argv, **kwarg)

        # do automatic bind
        # FIXME: unfortunately we can not omit it here
        try:
            self.bind(groups, async_cache=async_cache)
        except:
            # thanks to jtluka at redhat.com and the LNST
            # team for the fixed fd leak
            super(DL, self).close()
            raise

    def list(self):
        return self.get_dump()

    def get_dump(self):
        msg = devlinkcmd()
        msg['cmd'] = DEVLINK_NAMES['DEVLINK_CMD_GET']
        return self.nlm_request(
            msg, msg_type=self.prid, msg_flags=NLM_F_REQUEST | NLM_F_DUMP
        )

    def port_list(self):
        return self.get_port_dump()

    def get_port_dump(self):
        msg = devlinkcmd()
        msg['cmd'] = DEVLINK_NAMES['DEVLINK_CMD_PORT_GET']
        return self.nlm_request(
            msg, msg_type=self.prid, msg_flags=NLM_F_REQUEST | NLM_F_DUMP
        )



import ipaddress
import json
from collections import OrderedDict
from socket import AF_INET, AF_INET6

from pr2modules.common import AF_MPLS, dqn2int, get_address_family


class MPLSTarget(OrderedDict):
    def __init__(self, prime=None):
        super(MPLSTarget, self).__init__()
        if prime is None:
            prime = {}
        elif isinstance(prime, str):
            prime = {'label': int(prime)}
        elif isinstance(prime, int):
            prime = {'label': prime}
        elif isinstance(prime, dict):
            pass
        else:
            raise TypeError()
        self['label'] = prime.get('label', 16)
        self['tc'] = prime.get('tc', 0)
        self['bos'] = prime.get('bos', 1)
        self['ttl'] = prime.get('ttl', 0)

    def __eq__(self, right):
        return (
            isinstance(right, (dict, MPLSTarget))
            and self['label'] == right.get('label', 16)
            and self['tc'] == right.get('tc', 0)
            and self['bos'] == right.get('bos', 1)
            and self['ttl'] == right.get('ttl', 0)
        )

    def __repr__(self):
        return repr(dict(self))


class IPTargets:
    add_defaults = True

    def parse_target(self, key, context, value):
        ret = {key: value}
        set_full_mask = False
        if isinstance(value, int):
            return {'family': AF_MPLS, key: MPLSTarget(value)}
        if isinstance(value, (list, tuple)):
            targets = []
            target = None
            for spec in value:
                target = MPLSTarget(spec)
                target['bos'] = 0
                targets.append(target)
            if target:
                target['bos'] = 1
                return {'family': AF_MPLS, key: targets}
            else:
                return {}
        if isinstance(value, dict):
            if value.get('label'):
                return {'family': AF_MPLS, key: MPLSTarget(value)}
            # do not overwrite message family for IP VIA
            return {key: value}
        try:
            return self.parse_target(key, context, json.loads(value))
        except (json.JSONDecodeError, TypeError):
            pass
        if isinstance(value, str):
            labels = value.split('/')
            if len(labels) > 1:
                # MPLS label stack simple syntax? e.g.: 16/24, 200/300 etc.
                try:
                    return self.parse_target(key, context, labels)
                except ValueError:
                    pass
                # only simple IP targets are left
                value, prefixlen = labels
                ret[key] = value
                if self.add_defaults:
                    if prefixlen.find('.') > 0:
                        ret[f'{key}_len'] = dqn2int(prefixlen, AF_INET)
                    elif prefixlen.find(':') >= 0:
                        ret[f'{key}_len'] = dqn2int(prefixlen, AF_INET6)
                    else:
                        ret[f'{key}_len'] = int(prefixlen)
            else:
                if (
                    self.add_defaults
                    and key in ('dst', 'src')
                    and f'{key}_len' not in context
                ):
                    set_full_mask = True

            if self.add_defaults:
                ret['family'] = get_address_family(value)
                if ret['family'] == AF_INET6:
                    ret[key] = ipaddress.ip_address(value).compressed
                if set_full_mask:
                    if ret['family'] == AF_INET6:
                        ret[f'{key}_len'] = 128
                    elif ret['family'] == AF_INET:
                        ret[f'{key}_len'] = 32
        return ret

    def set_dst(self, context, value):
        if value in ('', 'default'):
            return {'dst': ''}
        elif value in ('0', '0.0.0.0'):
            return {'dst': '', 'family': AF_INET}
        elif value in ('::', '::/0'):
            return {'dst': '', 'family': AF_INET6}
        return self.parse_target('dst', context, value)

    def set_src(self, context, value):
        return self.parse_target('src', context, value)

    def set_via(self, context, value):
        return self.parse_target('via', context, value)

    def set_newdst(self, context, value):
        return self.parse_target('newdst', context, value)

    def set_gateway(self, context, value):
        return self.parse_target('gateway', context, value)


class Index:
    def set_index(self, context, value):
        if isinstance(value, (list, tuple)):
            value = value[0]
        return {'index': value}


class IPRouteFilter:
    def __init__(self, command):
        self.command = command


class NLAKeyTransform:

    _nla_prefix = ''

    def _key_transform(self, key):
        if key.startswith(self._nla_prefix):
            key = key[len(self._nla_prefix) :].lower()
        return key



'''
choke
+++++

Parameters:

    * `limit` (required) -- int
    * `bandwith` (required) -- str/int
    * `min` -- int
    * `max` -- int
    * `avpkt` -- str/int, packet size
    * `burst` -- int
    * `probability` -- float
    * `ecn` -- bool

Example::

    ip.tc('add', 'choke', interface,
          limit=5500,
          bandwith="10mbit",
          ecn=True)

'''
import logging
import struct

from pr2modules.netlink import nla, nla_string
from pr2modules.netlink.rtnl import TC_H_ROOT
from pr2modules.netlink.rtnl.tcmsg.common import (
    get_rate,
    get_size,
    red_eval_ewma,
    red_eval_idle_damping,
    red_eval_P,
)
from pr2modules.netlink.rtnl.tcmsg.common import stats2 as c_stats2

log = logging.getLogger(__name__)
parent = TC_H_ROOT


def get_parameters(kwarg):
    # The code is ported from iproute2
    avpkt = 1000
    probability = 0.02
    opt = {
        'limit': kwarg['limit'],  # required
        'qth_min': kwarg.get('min', 0),
        'qth_max': kwarg.get('max', 0),
        'Wlog': 0,
        'Plog': 0,
        'Scell_log': 0,
        'flags': 1 if kwarg.get('ecn') else 0,
    }

    rate = get_rate(kwarg['bandwith'])  # required
    burst = kwarg.get('burst', 0)
    avpkt = get_size(kwarg.get('avpkt', 1000))
    probability = kwarg.get('probability', 0.02)

    if not opt['qth_max']:
        opt['qth_max'] = opt['limit'] // 4
    if not opt['qth_min']:
        opt['qth_min'] = opt['qth_max'] // 3
    if not burst:
        burst = (2 * opt['qth_min'] + opt['qth_max']) // 3

    if opt['qth_max'] > opt['limit']:
        raise Exception('max is larger than limit')
    if opt['qth_min'] >= opt['qth_max']:
        raise Exception('min is not smaller than max')

    # Wlog
    opt['Wlog'] = red_eval_ewma(opt['qth_min'] * avpkt, burst, avpkt)
    if opt['Wlog'] < 0:
        raise Exception('failed to calculate EWMA')
    elif opt['Wlog'] > 10:
        log.warning('choke: burst %s seems to be too large' % burst)
    # Plog
    opt['Plog'] = red_eval_P(
        opt['qth_min'] * avpkt, opt['qth_max'] * avpkt, probability
    )
    if opt['Plog'] < 0:
        raise Exception('choke: failed to calculate probability')
    # Scell_log, stab
    opt['Scell_log'], stab = red_eval_idle_damping(opt['Wlog'], avpkt, rate)
    if opt['Scell_log'] < 0:
        raise Exception('choke: failed to calculate idle damping table')

    return {
        'attrs': [
            ['TCA_CHOKE_PARMS', opt],
            ['TCA_CHOKE_STAB', stab],
            ['TCA_CHOKE_MAX_P', int(probability * pow(2, 32))],
        ]
    }


class options(nla):
    nla_map = (
        ('TCA_CHOKE_UNSPEC', 'none'),
        ('TCA_CHOKE_PARMS', 'qopt'),
        ('TCA_CHOKE_STAB', 'stab'),
        ('TCA_CHOKE_MAX_P', 'uint32'),
    )

    class qopt(nla):
        fields = (
            ('limit', 'I'),
            ('qth_min', 'I'),
            ('qth_max', 'I'),
            ('Wlog', 'B'),
            ('Plog', 'B'),
            ('Scell_log', 'B'),
            ('flags', 'B'),
        )

    class stab(nla_string):
        def encode(self):
            self['value'] = struct.pack(
                'B' * 256, *(int(x) for x in self.value)
            )
            nla_string.encode(self)


class stats(nla):
    fields = (
        ('early', 'I'),
        ('pdrop', 'I'),
        ('other', 'I'),
        ('marked', 'I'),
        ('matched', 'I'),
    )


class stats2(c_stats2):
    class stats_app(stats):
        pass



import errno
import sys
import types

from pr2modules import config
from pr2modules.common import DEFAULT_RCVBUF, AddrPool, Namespace
from pr2modules.netlink import NETLINK_ROUTE, rtnl
from pr2modules.netlink.nlsocket import (
    BatchSocket,
    ChaoticNetlinkSocket,
    NetlinkSocket,
)
from pr2modules.netlink.rtnl.marshal import MarshalRtnl
from pr2modules.proxy import NetlinkProxy

if sys.platform.startswith('linux'):
    if config.kernel < [3, 3, 0]:
        from pr2modules.netlink.rtnl.ifinfmsg.compat import (
            proxy_dellink,
            proxy_linkinfo,
            proxy_newlink,
            proxy_setlink,
        )
    else:
        from pr2modules.netlink.rtnl.ifinfmsg.proxy import (
            proxy_newlink,
            proxy_setlink,
        )


class IPRSocketBase(object):
    def __init__(self, *argv, **kwarg):
        if 'family' in kwarg:
            kwarg.pop('family')
        super(IPRSocketBase, self).__init__(NETLINK_ROUTE, *argv[1:], **kwarg)
        self.marshal = MarshalRtnl()
        self._s_channel = None
        if sys.platform.startswith('linux'):
            self._gate = self._gate_linux
            self.sendto_gate = self._gate_linux
            send_ns = Namespace(
                self,
                {'addr_pool': AddrPool(0x10000, 0x1FFFF), 'monitor': False},
            )
            self._sproxy = NetlinkProxy(policy='return', nl=send_ns)
            self._sproxy.pmap = {
                rtnl.RTM_NEWLINK: proxy_newlink,
                rtnl.RTM_SETLINK: proxy_setlink,
            }
            if config.kernel < [3, 3, 0]:
                self._recv_ns = Namespace(
                    self,
                    {
                        'addr_pool': AddrPool(0x20000, 0x2FFFF),
                        'monitor': False,
                    },
                )
                self._sproxy.pmap[rtnl.RTM_DELLINK] = proxy_dellink
                # inject proxy hooks into recv() and...
                self.__recv = self._recv
                self._recv = self._p_recv
                # ... recv_into()
                self._recv_ft = self.recv_ft
                self.recv_ft = self._p_recv_ft

    def bind(self, groups=rtnl.RTMGRP_DEFAULTS, **kwarg):
        super(IPRSocketBase, self).bind(groups, **kwarg)

    def _gate_linux(self, msg, addr):
        msg.reset()
        msg.encode()
        ret = self._sproxy.handle(msg)
        if ret is not None:
            if ret['verdict'] == 'forward':
                return self._sendto(ret['data'], addr)
            elif ret['verdict'] in ('return', 'error'):
                if self._s_channel is not None:
                    return self._s_channel.send(ret['data'])
                else:
                    msgs = self.marshal.parse(ret['data'])
                    for msg in msgs:
                        seq = msg['header']['sequence_number']
                        if seq in self.backlog:
                            self.backlog[seq].append(msg)
                        else:
                            self.backlog[seq] = [msg]
                    return len(ret['data'])
            else:
                ValueError('Incorrect verdict')

        return self._sendto(msg.data, addr)

    def _p_recv_ft(self, bufsize, flags=0):
        data = self._recv_ft(bufsize, flags)
        ret = proxy_linkinfo(data, self._recv_ns)
        if ret is not None:
            if ret['verdict'] in ('forward', 'error'):
                return ret['data']
            else:
                ValueError('Incorrect verdict')

        return data

    def _p_recv(self, bufsize, flags=0):
        data = self.__recv(bufsize, flags)
        ret = proxy_linkinfo(data, self._recv_ns)
        if ret is not None:
            if ret['verdict'] in ('forward', 'error'):
                return ret['data']
            else:
                ValueError('Incorrect verdict')

        return data


class IPBatchSocket(IPRSocketBase, BatchSocket):
    pass


class ChaoticIPRSocket(IPRSocketBase, ChaoticNetlinkSocket):
    pass


class IPRSocket(IPRSocketBase, NetlinkSocket):
    '''
    The simplest class, that connects together the netlink parser and
    a generic Python socket implementation. Provides method get() to
    receive the next message from netlink socket and parse it. It is
    just simple socket-like class, it implements no buffering or
    like that. It spawns no additional threads, leaving this up to
    developers.

    Please note, that netlink is an asynchronous protocol with
    non-guaranteed delivery. You should be fast enough to get all the
    messages in time. If the message flow rate is higher than the
    speed you parse them with, exceeding messages will be dropped.

    *Usage*

    Threadless RT netlink monitoring with blocking I/O calls:

        >>> from pr2modules import IPRSocket
        >>> from pprint import pprint
        >>> s = IPRSocket()
        >>> s.bind()
        >>> pprint(s.get())
        [{'attrs': [('RTA_TABLE', 254),
                    ('RTA_DST', '2a00:1450:4009:808::1002'),
                    ('RTA_GATEWAY', 'fe80:52:0:2282::1fe'),
                    ('RTA_OIF', 2),
                    ('RTA_PRIORITY', 0),
                    ('RTA_CACHEINFO', {'rta_clntref': 0,
                                       'rta_error': 0,
                                       'rta_expires': 0,
                                       'rta_id': 0,
                                       'rta_lastuse': 5926,
                                       'rta_ts': 0,
                                       'rta_tsage': 0,
                                       'rta_used': 1})],
          'dst_len': 128,
          'event': 'RTM_DELROUTE',
          'family': 10,
          'flags': 512,
          'header': {'error': None,
                     'flags': 0,
                     'length': 128,
                     'pid': 0,
                     'sequence_number': 0,
                     'type': 25},
          'proto': 9,
          'scope': 0,
          'src_len': 0,
          'table': 254,
          'tos': 0,
          'type': 1}]
        >>>
    '''

    _brd_socket = None

    def bind(self, *argv, **kwarg):
        if kwarg.pop('clone_socket', False):
            self._brd_socket = self.clone()

            def get(
                self,
                bufsize=DEFAULT_RCVBUF,
                msg_seq=0,
                terminate=None,
                callback=None,
            ):
                if msg_seq == 0:
                    return self._brd_socket.get(
                        bufsize, msg_seq, terminate, callback
                    )
                else:
                    return super(IPRSocket, self).get(
                        bufsize, msg_seq, terminate, callback
                    )

            def close(self, code=errno.ECONNRESET):
                with self.sys_lock:
                    self._brd_socket.close()
                    return super(IPRSocket, self).close(code=code)

            self.get = types.MethodType(get, self)
            self.close = types.MethodType(close, self)
            kwarg['recursive'] = True
            return self._brd_socket.bind(*argv, **kwarg)
        else:
            return super(IPRSocket, self).bind(*argv, **kwarg)



import socket
import struct

from pr2modules import config
from pr2modules.common import hexdump
from pr2modules.netlink import nlmsg_base

if config.uname[0] == 'OpenBSD':
    from pr2modules.bsd.pf_route.openbsd import (
        IFF_NAMES,
        IFF_VALUES,
        bsdmsg,
        if_announcemsg,
        if_msg,
        ifa_msg_base,
        ifma_msg_base,
        rt_msg_base,
    )
else:
    from pr2modules.bsd.pf_route.freebsd import (
        IFF_NAMES,
        IFF_VALUES,
        bsdmsg,
        if_announcemsg,
        if_msg,
        ifa_msg_base,
        ifma_msg_base,
        rt_msg_base,
    )


RTAX_MAX = 8


class rt_slot(nlmsg_base):

    __slots__ = ()
    header = (('length', 'B'), ('family', 'B'))


class rt_msg(rt_msg_base):

    __slots__ = ()
    force_mask = False

    class hex(rt_slot):
        def decode(self):
            rt_slot.decode(self)
            length = self['header']['length']
            self['value'] = hexdump(
                self.data[self.offset + 2 : self.offset + length]
            )

    class rt_slot_ifp(rt_slot):
        def decode(self):
            rt_slot.decode(self)
            #
            # Structure
            #     0       1       2       3       4       5       6       7
            # |-------+-------+-------+-------|-------+-------+-------+-------|
            # |  len  |  fam  |    ifindex    |   ?   |  nlen |    padding?   |
            # |-------+-------+-------+-------|-------+-------+-------+-------|
            # | ...
            # | ...
            #
            # len -- sockaddr len
            # fam -- sockaddr family
            # ifindex -- interface index
            # ? -- no idea, probably again some sockaddr related info?
            # nlen -- device name length
            # padding? -- probably structure alignment
            #
            (self['index'], _, name_length) = struct.unpack(
                'HBB', self.data[self.offset + 2 : self.offset + 6]
            )
            self['ifname'] = self.data[
                self.offset + 8 : self.offset + 8 + name_length
            ]

    class rt_slot_addr(rt_slot):
        def decode(self):
            alen = {socket.AF_INET: 4, socket.AF_INET6: 16}
            rt_slot.decode(self)
            #
            # Yksinkertainen: only the sockaddr family (one byte) and the
            # network address.
            #
            # But for netmask it's completely screwed up. E.g.:
            #
            #  ifconfig disc2 10.0.0.1 255.255.255.0 up
            # -->
            #  ... NETMASK: 38:12:00:00:ff:00:00:00:00:00:00:...
            #
            # Why?!
            #
            family = self['header']['family']
            length = self['header']['length']
            if family in (socket.AF_INET, socket.AF_INET6):
                addrlen = alen.get(family, 0)
                data = self.data[self.offset + 4 : self.offset + 4 + addrlen]
                self['address'] = socket.inet_ntop(family, data)
            else:
                # FreeBSD and OpenBSD use different approaches
                # FreeBSD: family == 0x12
                # OpenBSD: family == 0x0
                if self.parent.force_mask and family in (0x0, 0x12):
                    data = self.data[self.offset + 4 : self.offset + 8]
                    data = data + b'\0' * (4 - len(data))
                    self['address'] = socket.inet_ntop(socket.AF_INET, data)
                else:
                    self['raw'] = self.data[self.offset : self.offset + length]

    def decode(self):
        bsdmsg.decode(self)
        offset = self.sockaddr_offset
        for i in range(RTAX_MAX):
            if self['rtm_addrs'] & (1 << i):
                handler = getattr(self, self.ifa_slots[i][1])
                slot = handler(self.data[offset:], parent=self)
                slot.decode()
                offset += slot['header']['length']
                self[self.ifa_slots[i][0]] = slot


class ifa_msg(ifa_msg_base, rt_msg):
    force_mask = True


class ifma_msg(ifma_msg_base, rt_msg):
    pass


__all__ = (
    bsdmsg,
    if_msg,
    rt_msg,
    ifa_msg,
    ifma_msg,
    if_announcemsg,
    IFF_NAMES,
    IFF_VALUES,
)



from pr2modules.netlink import nla, nlmsg


class ndtmsg(nlmsg):
    '''
    Neighbour table message
    '''

    __slots__ = ()

    fields = (('family', 'B'), ('__pad', '3x'))

    nla_map = (
        ('NDTA_UNSPEC', 'none'),
        ('NDTA_NAME', 'asciiz'),
        ('NDTA_THRESH1', 'uint32'),
        ('NDTA_THRESH2', 'uint32'),
        ('NDTA_THRESH3', 'uint32'),
        ('NDTA_CONFIG', 'config'),
        ('NDTA_PARMS', 'parms'),
        ('NDTA_STATS', 'stats'),
        ('NDTA_GC_INTERVAL', 'uint64'),
    )

    class config(nla):

        __slots__ = ()

        fields = (
            ('key_len', 'H'),
            ('entry_size', 'H'),
            ('entries', 'I'),
            ('last_flush', 'I'),  # delta to now in msecs
            ('last_rand', 'I'),  # delta to now in msecs
            ('hash_rnd', 'I'),
            ('hash_mask', 'I'),
            ('hash_chain_gc', 'I'),
            ('proxy_qlen', 'I'),
        )

    class stats(nla):

        __slots__ = ()

        fields = (
            ('allocs', 'Q'),
            ('destroys', 'Q'),
            ('hash_grows', 'Q'),
            ('res_failed', 'Q'),
            ('lookups', 'Q'),
            ('hits', 'Q'),
            ('rcv_probes_mcast', 'Q'),
            ('rcv_probes_ucast', 'Q'),
            ('periodic_gc_runs', 'Q'),
            ('forced_gc_runs', 'Q'),
        )

    class parms(nla):

        __slots__ = ()

        nla_map = (
            ('NDTPA_UNSPEC', 'none'),
            ('NDTPA_IFINDEX', 'uint32'),
            ('NDTPA_REFCNT', 'uint32'),
            ('NDTPA_REACHABLE_TIME', 'uint64'),
            ('NDTPA_BASE_REACHABLE_TIME', 'uint64'),
            ('NDTPA_RETRANS_TIME', 'uint64'),
            ('NDTPA_GC_STALETIME', 'uint64'),
            ('NDTPA_DELAY_PROBE_TIME', 'uint64'),
            ('NDTPA_QUEUE_LEN', 'uint32'),
            ('NDTPA_APP_PROBES', 'uint32'),
            ('NDTPA_UCAST_PROBES', 'uint32'),
            ('NDTPA_MCAST_PROBES', 'uint32'),
            ('NDTPA_ANYCAST_DELAY', 'uint64'),
            ('NDTPA_PROXY_DELAY', 'uint64'),
            ('NDTPA_PROXY_QLEN', 'uint32'),
            ('NDTPA_LOCKTIME', 'uint64'),
            ('NDTPA_QUEUE_LENBYTES', 'uint32'),
        )



from pr2modules.netlink import nla


class ipvlan(nla):
    prefix = 'IFLA_'
    nla_map = (('IFLA_IPVLAN_UNSPEC', 'none'), ('IFLA_IPVLAN_MODE', 'uint16'))

    modes = {
        0: 'IPVLAN_MODE_L2',
        1: 'IPVLAN_MODE_L3',
        'IPVLAN_MODE_L2': 0,
        'IPVLAN_MODE_L3': 1,
    }



from pr2modules.netlink import nlmsg, nlmsg_atoms


class nsinfmsg(nlmsg):
    '''
    Fake message type to represent network namespace information.

    This is a prototype, the NLA layout is subject to change without
    notification.
    '''

    __slots__ = ()
    prefix = 'NSINFO_'

    fields = (('inode', 'I'), ('netnsid', 'I'))

    nla_map = (
        ('NSINFO_UNSPEC', 'none'),
        ('NSINFO_PATH', 'string'),
        ('NSINFO_PEER', 'peer'),
    )

    class peer(nlmsg_atoms.string):
        sql_type = None



from pr2modules.netlink.rtnl import TC_H_ROOT
from pr2modules.netlink.rtnl.tcmsg.common import (
    get_rate_parameters,
    nla_plus_rtab,
)

parent = TC_H_ROOT


def get_parameters(kwarg):
    parms = get_rate_parameters(kwarg)
    # fill parameters
    return {'attrs': [['TCA_TBF_PARMS', parms], ['TCA_TBF_RTAB', True]]}


class options(nla_plus_rtab):
    nla_map = (
        ('TCA_TBF_UNSPEC', 'none'),
        ('TCA_TBF_PARMS', 'tbf_parms'),
        ('TCA_TBF_RTAB', 'rtab'),
        ('TCA_TBF_PTAB', 'ptab'),
    )

    class tbf_parms(nla_plus_rtab.parms):
        fields = (
            ('rate_cell_log', 'B'),
            ('rate___reserved', 'B'),
            ('rate_overhead', 'H'),
            ('rate_cell_align', 'h'),
            ('rate_mpu', 'H'),
            ('rate', 'I'),
            ('peak_cell_log', 'B'),
            ('peak___reserved', 'B'),
            ('peak_overhead', 'H'),
            ('peak_cell_align', 'h'),
            ('peak_mpu', 'H'),
            ('peak', 'I'),
            ('limit', 'I'),
            ('buffer', 'I'),
            ('mtu', 'I'),
        )



from pr2modules.netlink import nla, nlmsg
from pr2modules.netlink.rtnl.ifinfmsg import ifinfmsg


class ifstatsmsg(nlmsg):
    fields = (
        ('family', 'B'),
        ('pad1', 'B'),
        ('pad2', 'H'),
        ('ifindex', 'I'),
        ('filter_mask', 'I'),
    )

    nla_map = (
        ('IFLA_STATS_UNSPEC', 'none'),
        ('IFLA_STATS_LINK_64', 'ifstats64'),
        ('IFLA_STATS_LINK_XSTATS', 'ifxstats'),
        ('IFLA_STATS_LINK_XSTATS_SLAVE', 'ifxstats'),
        ('IFLA_STATS_LINK_OFFLOAD_XSTATS', 'hex'),
        ('IFLA_STATS_AF_SPEC', 'hex'),
    )

    class ifstats64(ifinfmsg.ifstats64):
        pass

    class ifxstats(nla):
        nla_map = (
            ('LINK_XSTATS_TYPE_UNSPEC', 'none'),
            ('LINK_XSTATS_TYPE_BRIDGE', 'bridge'),
            ('LINK_XSTATS_TYPE_BOND', 'hex'),
        )

        class bridge(nla):
            nla_map = (
                ('BRIDGE_XSTATS_UNSPEC', 'none'),
                ('BRIDGE_XSTATS_VLAN', 'vlan'),
                ('BRIDGE_XSTATS_MCAST', 'mcast'),
                ('BRIDGE_XSTATS_PAD', 'hex'),
                ('BRIDGE_XSTATS_STP', 'stp'),
            )

            class vlan(nla):
                fields = (
                    ('rx_bytes', 'Q'),
                    ('rx_packets', 'Q'),
                    ('tx_bytes', 'Q'),
                    ('tx_packets', 'Q'),
                    ('vid', 'H'),
                    ('flags', 'H'),
                    ('pad2', 'I'),
                )

            class mcast(nla):
                fields = (
                    ('igmp_v1queries', 'QQ'),
                    ('igmp_v2queries', 'QQ'),
                    ('igmp_v3queries', 'QQ'),
                    ('igmp_leaves', 'QQ'),
                    ('igmp_v1reports', 'QQ'),
                    ('igmp_v2reports', 'QQ'),
                    ('igmp_v3reports', 'QQ'),
                    ('igmp_parse_errors', 'Q'),
                    ('mld_v1queries', 'QQ'),
                    ('mld_v2queries', 'QQ'),
                    ('mld_leaves', 'QQ'),
                    ('mld_v1reports', 'QQ'),
                    ('mld_v2reports', 'QQ'),
                    ('mld_parse_errors', 'Q'),
                    ('mcast_bytes', 'QQ'),
                    ('mcast_packets', 'QQ'),
                )

            class stp(nla):
                fields = (
                    ('transition_blk', 'Q'),
                    ('transition_fwd', 'Q'),
                    ('rx_bpdu', 'Q'),
                    ('tx_bpdu', 'Q'),
                    ('rx_tcn', 'Q'),
                    ('tx_tcn', 'Q'),
                )



from pr2modules.netlink import NETLINK_ROUTE, rtnl
from pr2modules.netlink.nlsocket import NetlinkSocket
from pr2modules.netlink.rtnl.marshal import MarshalRtnl


class RawIPRSocketBase(object):
    def __init__(self, fileno=None):
        super(RawIPRSocketBase, self).__init__(NETLINK_ROUTE, fileno=fileno)
        self.marshal = MarshalRtnl()

    def bind(self, groups=rtnl.RTMGRP_DEFAULTS, **kwarg):
        super(RawIPRSocketBase, self).bind(groups, **kwarg)


class RawIPRSocket(RawIPRSocketBase, NetlinkSocket):
    pass



'''
clsact
++++++

The clsact qdisc provides a mechanism to attach integrated
filter-action classifiers to an interface, either at ingress or egress,
or both. The use case shown here is using a bpf program (implemented
elsewhere) to direct the packet processing. The example also uses the
direct-action feature to specify what to do with each packet (pass,
drop, redirect, etc.).

BPF ingress/egress example using clsact qdisc::

    # open_bpf_fd is outside the scope of pyroute2
    #fd = open_bpf_fd()
    eth0 = ip.get_links(ifname="eth0")[0]
    ip.tc("add", "clsact", eth0)
    # add ingress clsact
    ip.tc("add-filter", "bpf", idx, ":1", fd=fd, name="myprog",
          parent="ffff:fff2", classid=1, direct_action=True)
    # add egress clsact
    ip.tc("add-filter", "bpf", idx, ":1", fd=fd, name="myprog",
          parent="ffff:fff3", classid=1, direct_action=True)

'''
from pr2modules.netlink.rtnl import TC_H_CLSACT

parent = TC_H_CLSACT


def fix_msg(msg, kwarg):
    msg['handle'] = 0xFFFF0000



import multiprocessing
import platform
import re
import signal
import socket

kernel_version_re = re.compile('^[0-9.]+')


def parse_kernel_version(kernel_name):
    match_obj = kernel_version_re.match(kernel_name)
    if match_obj is None:
        return []
    return [int(x) for x in kernel_name[0 : match_obj.end()].split('.') if x]


SocketBase = socket.socket
MpPipe = multiprocessing.Pipe
MpQueue = multiprocessing.Queue
MpProcess = multiprocessing.Process
ipdb_nl_async = True
nlm_generator = False
nla_via_getattr = False
async_qsize = 4096
commit_barrier = 0
gc_timeout = 60
db_transaction_limit = 1
cache_expire = 60

signal_stop_remote = signal.SIGUSR1

# save uname() on startup time: it is not so
# highly possible that the kernel will be
# changed in runtime, while calling uname()
# every time is a bit expensive
uname = tuple(platform.uname())
machine = platform.machine()
arch = platform.architecture()[0]
kernel = parse_kernel_version(uname[2])

AF_BRIDGE = getattr(socket, 'AF_BRIDGE', 7)
AF_NETLINK = getattr(socket, 'AF_NETLINK', 16)

data_plugins_pkgs = []
data_plugins_path = []

netns_path = ['/var/run/netns', '/var/run/docker/netns']

entry_points_aliases = {
    'pr2modules.netlink.exceptions': 'pyroute2.netlink.exceptions'
}



import os
import select
import threading

from pr2modules.netlink.rtnl import RTM_VALUES
from pr2modules.netlink.rtnl.riprsocket import RawIPRSocket


def sync(f):
    '''
    A decorator to wrap up external utility calls.

    A decorated function receives a netlink message
    as a parameter, and then:

    1. Starts a monitoring thread
    2. Performs the external call
    3. Waits for a netlink event specified by `msg`
    4. Joins the monitoring thread

    If the wrapped function raises an exception, the
    monitoring thread will be forced to stop via the
    control channel pipe. The exception will be then
    forwarded.
    '''

    def monitor(event, ifname, cmd):
        with RawIPRSocket() as ipr:
            poll = select.poll()
            poll.register(ipr, select.POLLIN | select.POLLPRI)
            poll.register(cmd, select.POLLIN | select.POLLPRI)
            ipr.bind()
            while True:
                events = poll.poll()
                for (fd, event) in events:
                    if fd == ipr.fileno():
                        msgs = ipr.get()
                        for msg in msgs:
                            if (
                                msg.get('event') == event
                                and msg.get_attr('IFLA_IFNAME') == ifname
                            ):
                                return
                    else:
                        return

    def decorated(msg):
        rcmd, cmd = os.pipe()
        t = threading.Thread(
            target=monitor,
            args=(
                RTM_VALUES[msg['header']['type']],
                msg.get_attr('IFLA_IFNAME'),
                rcmd,
            ),
        )
        t.start()
        ret = None
        try:
            ret = f(msg)
        except Exception:
            raise
        finally:
            os.write(cmd, b'q')
            t.join()
            os.close(rcmd)
            os.close(cmd)
        return ret

    return decorated



from pr2modules.common import AF_MPLS
from pr2modules.netlink.rtnl import encap_type, rt_proto, rt_scope, rt_type
from pr2modules.netlink.rtnl.rtmsg import LWTUNNEL_ENCAP_MPLS
from pr2modules.netlink.rtnl.rtmsg import nh as nh_header
from pr2modules.netlink.rtnl.rtmsg import rtmsg

from .common import IPRouteFilter, IPTargets, MPLSTarget, NLAKeyTransform

encap_types = {'mpls': 1, AF_MPLS: 1, 'seg6': 5, 'bpf': 6, 'seg6local': 7}


class RouteFieldFilter(IPTargets, NLAKeyTransform):

    _nla_prefix = 'RTA_'

    def __init__(self, add_defaults=True):
        self.add_defaults = add_defaults

    def index(self, key, context, value):
        if isinstance(value, (list, tuple)):
            value = value[0]
        return {key: value}

    def set_oif(self, context, value):
        return self.index('oif', context, value)

    def set_iif(self, context, value):
        return self.index('iif', context, value)

    def set_family(self, context, value):
        if value == AF_MPLS:
            return {'family': AF_MPLS, 'dst_len': 20, 'table': 254, 'type': 1}
        return {'family': value}

    def set_flags(self, context, value):
        if context.get('family') == AF_MPLS:
            return {}
        if isinstance(value, (list, tuple, str)):
            return {'flags': rtmsg.names2flags(value)}
        return {'flags': value}

    def set_encap(self, context, value):
        # FIXME: planned for the next refactoring cycle
        if isinstance(value, dict) and value.get('type') == 'mpls':
            na = []
            target = None
            labels = value.get('labels', [])
            if isinstance(labels, (dict, int)):
                labels = [labels]
            if isinstance(labels, str):
                labels = labels.split('/')
            for label in labels:
                target = MPLSTarget(label)
                target['bos'] = 0
                na.append(target)
            target['bos'] = 1
            return {'encap_type': LWTUNNEL_ENCAP_MPLS, 'encap': na}
        return {'encap': value}

    def set_scope(self, context, value):
        if isinstance(value, str):
            return {'scope': rt_scope[value]}
        return {'scope': value}

    def set_proto(self, context, value):
        if isinstance(value, str):
            return {'proto': rt_proto[value]}
        return {'proto': value}

    def set_encap_type(self, context, value):
        if isinstance(value, str):
            return {'encap_type': encap_type[value]}
        return {'encap_type': value}

    def set_type(self, context, value):
        if isinstance(value, str):
            return {'type': rt_type[value]}
        return {'type': value}


class RouteIPRouteFilter(IPRouteFilter):
    def set_metrics(self, context, value):
        if value and 'attrs' not in value:
            metrics = {'attrs': []}
            for name, metric in value.items():
                rtax = rtmsg.metrics.name2nla(name)
                if metric is not None:
                    metrics['attrs'].append([rtax, metric])
            if metrics['attrs']:
                return {'metrics': metrics}
        return {}

    def set_multipath(self, context, value):
        if value:
            ret = []
            for v in value:
                if 'attrs' in v:
                    ret.append(v)
                    continue
                nh = {'attrs': []}
                nh_fields = [x[0] for x in nh_header.fields]
                for name in nh_fields:
                    nh[name] = v.get(name, 0)
                for name in v:
                    if name in nh_fields or v[name] is None:
                        continue
                    if name == 'encap' and isinstance(v[name], dict):
                        if (
                            v[name].get('type', None) is None
                            or v[name].get('labels', None) is None
                        ):
                            continue
                        nh['attrs'].append(
                            [
                                'RTA_ENCAP_TYPE',
                                encap_types.get(
                                    v[name]['type'], v[name]['type']
                                ),
                            ]
                        )
                        nh['attrs'].append(
                            ['RTA_ENCAP', self.encap_header(v[name])]
                        )
                    elif name == 'newdst':
                        nh['attrs'].append(
                            ['RTA_NEWDST', self.mpls_rta(v[name])]
                        )
                    else:
                        rta = rtmsg.name2nla(name)
                        nh['attrs'].append([rta, v[name]])
                ret.append(nh)
            if ret:
                return {'multipath': ret}
        return {}

    def set_encap(self, context, value):
        if (
            isinstance(value, (list, tuple))
            and context.get('encap_type') == LWTUNNEL_ENCAP_MPLS
        ):
            return {'encap': {'attrs': [['MPLS_IPTUNNEL_DST', value]]}}
        elif isinstance(value, dict):
            # human-friendly form:
            #
            # 'encap': {'type': 'mpls',
            #           'labels': '200/300'}
            #
            # 'type' is mandatory
            if 'type' in value and 'labels' in value:
                return {
                    'encap_type': encap_types.get(
                        value['type'], value['type']
                    ),
                    'encap': self.encap_header(value),
                }
            # human-friendly form:
            #
            # 'encap': {'type': 'seg6',
            #           'mode': 'encap'
            #           'segs': '2000::5,2000::6'}
            #
            # 'encap': {'type': 'seg6',
            #           'mode': 'inline'
            #           'segs': '2000::5,2000::6'
            #           'hmac': 1}
            #
            # 'encap': {'type': 'seg6',
            #           'mode': 'encap'
            #           'segs': '2000::5,2000::6'
            #           'hmac': 0xf}
            #
            # 'encap': {'type': 'seg6',
            #           'mode': 'inline'
            #           'segs': ['2000::5', '2000::6']}
            #
            # 'type', 'mode' and 'segs' are mandatory
            if 'type' in value and 'mode' in value and 'segs' in value:
                return {
                    'encap_type': encap_types.get(
                        value['type'], value['type']
                    ),
                    'encap': self.encap_header(value),
                }
            elif 'type' in value and (
                'in' in value or 'out' in value or 'xmit' in value
            ):
                return {
                    'encap_type': encap_types.get(
                        value['type'], value['type']
                    ),
                    'encap': self.encap_header(value),
                }
            # human-friendly form:
            #
            # 'encap': {'type': 'seg6local',
            #           'action': 'End'}
            #
            # 'encap': {'type': 'seg6local',
            #           'action': 'End.DT6',
            #           'table': '10'}
            #
            # 'encap': {'type': 'seg6local',
            #           'action': 'End.DT4',
            #           'vrf_table': 10}
            #
            # 'encap': {'type': 'seg6local',
            #           'action': 'End.DX6',
            #           'nh6': '2000::5'}
            #
            # 'encap': {'type': 'seg6local',
            #           'action': 'End.B6'
            #           'srh': {'segs': '2000::5,2000::6',
            #                   'hmac': 0xf}}
            #
            # 'type' and 'action' are mandatory
            elif 'type' in value and 'action' in value:
                return {
                    'encap_type': encap_types.get(
                        value['type'], value['type']
                    ),
                    'encap': self.encap_header(value),
                }
        return {}

    def finalize(self, context):
        for key in context:
            if context[key] in ('', None):
                try:
                    del context[key]
                except KeyError:
                    pass

    def mpls_rta(self, value):
        # FIXME: planned for the next refactoring cycle
        ret = []
        if not isinstance(value, (list, tuple, set)):
            value = (value,)
        for label in value:
            ret.append(MPLSTarget(label))
        if ret:
            ret[-1]['bos'] = 1
        return ret

    def encap_header(self, header):
        '''
        Encap header transform. Format samples:

            {'type': 'mpls',
             'labels': '200/300'}

            {'type': AF_MPLS,
             'labels': (200, 300)}

            {'type': 'mpls',
             'labels': 200}

            {'type': AF_MPLS,
             'labels': [{'bos': 0, 'label': 200, 'ttl': 16},
                        {'bos': 1, 'label': 300, 'ttl': 16}]}
        '''
        if isinstance(header['type'], int) or (
            header['type'] in ('mpls', AF_MPLS, LWTUNNEL_ENCAP_MPLS)
        ):
            ret = []
            override_bos = True
            labels = header['labels']
            if isinstance(labels, str):
                labels = labels.split('/')
            if not isinstance(labels, (tuple, list, set)):
                labels = (labels,)
            for label in labels:
                if isinstance(label, dict):
                    # dicts append intact
                    override_bos = False
                    ret.append(label)
                else:
                    # otherwise construct label dict
                    if isinstance(label, str):
                        label = int(label)
                    ret.append({'bos': 0, 'label': label})
            # the last label becomes bottom-of-stack
            if override_bos:
                ret[-1]['bos'] = 1
            return {'attrs': [['MPLS_IPTUNNEL_DST', ret]]}
        '''
        Seg6 encap header transform. Format samples:

            {'type': 'seg6',
             'mode': 'encap',
             'segs': '2000::5,2000::6'}

            {'type': 'seg6',
             'mode': 'encap'
             'segs': '2000::5,2000::6',
             'hmac': 1}
        '''
        if header['type'] == 'seg6':
            # Init step
            ret = {}
            # Parse segs
            segs = header['segs']
            # If they are in the form in_addr6,in_addr6
            if isinstance(segs, str):
                # Create an array with the splitted values
                temp = segs.split(',')
                # Init segs
                segs = []
                # Iterate over the values
                for seg in temp:
                    # Discard empty string
                    if seg != '':
                        # Add seg to segs
                        segs.append(seg)
            # Retrieve mode
            mode = header['mode']
            # hmac is optional and contains the hmac key
            hmac = header.get('hmac', None)
            # Construct the new object
            ret = {'mode': mode, 'segs': segs}
            # If hmac is present convert to u32
            if hmac:
                # Add to ret the hmac key
                ret['hmac'] = hmac & 0xFFFFFFFF
            # Done return the object
            return {'attrs': [['SEG6_IPTUNNEL_SRH', ret]]}
        '''
        BPF encap header transform. Format samples:

            {'type': 'bpf',
             'in': {'fd':4, 'name':'firewall'}}

            {'type': 'bpf',
             'in'  : {'fd':4, 'name':'firewall'},
             'out' : {'fd':5, 'name':'stats'},
             'xmit': {'fd':6, 'name':'vlan_push', 'headroom':4}}
        '''
        if header['type'] == 'bpf':
            attrs = {}
            for key, value in header.items():
                if key not in ['in', 'out', 'xmit']:
                    continue

                obj = [
                    ['LWT_BPF_PROG_FD', value['fd']],
                    ['LWT_BPF_PROG_NAME', value['name']],
                ]
                if key == 'in':
                    attrs['LWT_BPF_IN'] = {'attrs': obj}
                elif key == 'out':
                    attrs['LWT_BPF_OUT'] = {'attrs': obj}
                elif key == 'xmit':
                    attrs['LWT_BPF_XMIT'] = {'attrs': obj}
                    if 'headroom' in value:
                        attrs['LWT_BPF_XMIT_HEADROOM'] = value['headroom']

            return {'attrs': list(attrs.items())}
        '''
        Seg6 encap header transform. Format samples:

            {'type': 'seg6local',
             'action': 'End.DT6',
             'table': '10'}

            {'type': 'seg6local',
             'action': 'End.DT4',
             'vrf_table': 10}

            {'type': 'seg6local',
             'action': 'End.B6',
             'table': '10'
             'srh': {'segs': '2000::5,2000::6'}}
        '''
        if header['type'] == 'seg6local':
            # Init step
            ret = {}
            table = None
            nh4 = None
            nh6 = None
            iif = None  # Actually not used
            oif = None
            srh = {}
            segs = []
            hmac = None
            prog_fd = None
            prog_name = None
            vrf_table = None
            # Parse segs
            if srh:
                segs = header['srh']['segs']
                # If they are in the form in_addr6,in_addr6
                if isinstance(segs, str):
                    # Create an array with the splitted values
                    temp = segs.split(',')
                    # Init segs
                    segs = []
                    # Iterate over the values
                    for seg in temp:
                        # Discard empty string
                        if seg != '':
                            # Add seg to segs
                            segs.append(seg)
                # hmac is optional and contains the hmac key
                hmac = header.get('hmac', None)
            # Retrieve action
            action = header['action']
            if action == 'End.X':
                # Retrieve nh6
                nh6 = header['nh6']
            elif action == 'End.T':
                # Retrieve table and convert to u32
                table = header['table'] & 0xFFFFFFFF
            elif action == 'End.DX2':
                # Retrieve oif and convert to u32
                oif = header['oif'] & 0xFFFFFFFF
            elif action == 'End.DX6':
                # Retrieve nh6
                nh6 = header['nh6']
            elif action == 'End.DX4':
                # Retrieve nh6
                nh4 = header['nh4']
            elif action == 'End.DT6':
                # Retrieve table
                table = header['table']
            elif action == 'End.DT4':
                # Retrieve vrf_table
                vrf_table = header['vrf_table']
            elif action == 'End.B6':
                # Parse segs
                segs = header['srh']['segs']
                # If they are in the form in_addr6,in_addr6
                if isinstance(segs, str):
                    # Create an array with the splitted values
                    temp = segs.split(',')
                    # Init segs
                    segs = []
                    # Iterate over the values
                    for seg in temp:
                        # Discard empty string
                        if seg != '':
                            # Add seg to segs
                            segs.append(seg)
                # hmac is optional and contains the hmac key
                hmac = header.get('hmac', None)
                srh['segs'] = segs
                # If hmac is present convert to u32
                if hmac:
                    # Add to ret the hmac key
                    srh['hmac'] = hmac & 0xFFFFFFFF
                srh['mode'] = 'inline'
            elif action == 'End.B6.Encaps':
                # Parse segs
                segs = header['srh']['segs']
                # If they are in the form in_addr6,in_addr6
                if isinstance(segs, str):
                    # Create an array with the splitted values
                    temp = segs.split(',')
                    # Init segs
                    segs = []
                    # Iterate over the values
                    for seg in temp:
                        # Discard empty string
                        if seg != '':
                            # Add seg to segs
                            segs.append(seg)
                # hmac is optional and contains the hmac key
                hmac = header.get('hmac', None)
                srh['segs'] = segs
                if hmac:
                    # Add to ret the hmac key
                    srh['hmac'] = hmac & 0xFFFFFFFF
                srh['mode'] = 'encap'
            elif action == 'End.BPF':
                prog_fd = header['bpf']['fd']
                prog_name = header['bpf']['name']

            # Construct the new object
            ret = []
            ret.append(['SEG6_LOCAL_ACTION', {'value': action}])
            if table:
                # Add the table to ret
                ret.append(['SEG6_LOCAL_TABLE', {'value': table}])
            if vrf_table:
                # Add the vrf_table to ret
                ret.append(['SEG6_LOCAL_VRFTABLE', {'value': vrf_table}])
            if nh4:
                # Add the nh4 to ret
                ret.append(['SEG6_LOCAL_NH4', {'value': nh4}])
            if nh6:
                # Add the nh6 to ret
                ret.append(['SEG6_LOCAL_NH6', {'value': nh6}])
            if iif:
                # Add the iif to ret
                ret.append(['SEG6_LOCAL_IIF', {'value': iif}])
            if oif:
                # Add the oif to ret
                ret.append(['SEG6_LOCAL_OIF', {'value': oif}])
            if srh:
                # Add the srh to ret
                ret.append(['SEG6_LOCAL_SRH', srh])
            if prog_fd and prog_name:
                # Add the prog_fd and prog_name to ret
                ret.append(
                    [
                        'SEG6_LOCAL_BPF',
                        {
                            'attrs': [
                                ['LWT_BPF_PROG_FD', prog_fd],
                                ['LWT_BPF_PROG_NAME', prog_name],
                            ]
                        },
                    ]
                )
            # Done return the object
            return {'attrs': ret}



from socket import htons

from pr2modules.netlink import nla
from pr2modules.netlink.rtnl.tcmsg.common import tc_actions

v_actions = {'pop': 1, 'push': 2, 'modify': 3}


class options(nla):
    nla_map = (
        ('TCA_VLAN_UNSPEC', 'none'),
        ('TCA_VLAN_TM', 'none'),
        ('TCA_VLAN_PARMS', 'tca_vlan_parms'),
        ('TCA_VLAN_PUSH_VLAN_ID', 'uint16'),
        ('TCA_VLAN_PUSH_VLAN_PROTOCOL', 'uint16'),
        ('TCA_VLAN_PAD', 'none'),
        ('TCA_VLAN_PUSH_VLAN_PRIORITY', 'uint8'),
    )

    class tca_vlan_parms(nla):
        fields = (
            ('index', 'I'),
            ('capab', 'I'),
            ('action', 'i'),
            ('refcnt', 'i'),
            ('bindcnt', 'i'),
            ('v_action', 'i'),
        )


def get_parameters(kwarg):
    ret = {'attrs': []}
    parms = {'v_action': v_actions[kwarg['v_action']]}
    parms['action'] = tc_actions[kwarg.get('action', 'pipe')]
    ret['attrs'].append(['TCA_VLAN_PARMS', parms])
    # Vlan id compulsory for "push" and "modify"
    if kwarg['v_action'] in ['push', 'modify']:
        ret['attrs'].append(['TCA_VLAN_PUSH_VLAN_ID', kwarg['id']])
    if 'priority' in kwarg:
        ret['attrs'].append(['TCA_VLAN_PUSH_VLAN_PRIORITY', kwarg['priority']])
    if kwarg.get('protocol', '802.1Q') == '802.1ad':
        ret['attrs'].append(['TCA_VLAN_PUSH_VLAN_PROTOCOL', htons(0x88A8)])
    else:
        ret['attrs'].append(['TCA_VLAN_PUSH_VLAN_PROTOCOL', htons(0x8100)])
    return ret



import socket

from pr2modules.netlink.nfnetlink.nfctsocket import (
    IP_CT_TCP_FLAG_TO_NAME,
    IPSBIT_TO_NAME,
    TCPF_TO_NAME,
    NFCTAttrTuple,
    NFCTSocket,
)


class NFCTATcpProtoInfo(object):

    __slots__ = (
        'state',
        'wscale_orig',
        'wscale_reply',
        'flags_orig',
        'flags_reply',
    )

    def __init__(
        self,
        state,
        wscale_orig=None,
        wscale_reply=None,
        flags_orig=None,
        flags_reply=None,
    ):
        self.state = state
        self.wscale_orig = wscale_orig
        self.wscale_reply = wscale_reply
        self.flags_orig = flags_orig
        self.flags_reply = flags_reply

    def state_name(self):
        return ','.join(
            [name for bit, name in TCPF_TO_NAME.items() if self.state & bit]
        )

    def flags_name(self, flags):
        if flags is None:
            return ''
        s = ''
        for bit, name in IP_CT_TCP_FLAG_TO_NAME.items():
            if flags & bit:
                s += '{},'.format(name)
        return s[:-1]

    @classmethod
    def from_netlink(cls, ndmsg):
        cta_tcp = ndmsg.get_attr('CTA_PROTOINFO_TCP')
        state = cta_tcp.get_attr('CTA_PROTOINFO_TCP_STATE')

        # second argument is the mask returned by kernel but useless for us
        flags_orig, _ = cta_tcp.get_attr('CTA_PROTOINFO_TCP_FLAGS_ORIGINAL')
        flags_reply, _ = cta_tcp.get_attr('CTA_PROTOINFO_TCP_FLAGS_REPLY')
        return cls(state=state, flags_orig=flags_orig, flags_reply=flags_reply)

    def __repr__(self):
        return 'TcpInfo(state={}, orig_flags={}, reply_flags={})'.format(
            self.state_name(),
            self.flags_name(self.flags_orig),
            self.flags_name(self.flags_reply),
        )


class ConntrackEntry(object):

    __slots__ = (
        'tuple_orig',
        'tuple_reply',
        'status',
        'timeout',
        'protoinfo',
        'mark',
        'id',
        'use',
    )

    def __init__(
        self,
        family,
        tuple_orig,
        tuple_reply,
        cta_status,
        cta_timeout,
        cta_protoinfo,
        cta_mark,
        cta_id,
        cta_use,
    ):
        self.tuple_orig = NFCTAttrTuple.from_netlink(family, tuple_orig)
        self.tuple_reply = NFCTAttrTuple.from_netlink(family, tuple_reply)

        self.status = cta_status
        self.timeout = cta_timeout

        if self.tuple_orig.proto == socket.IPPROTO_TCP:
            self.protoinfo = NFCTATcpProtoInfo.from_netlink(cta_protoinfo)
        else:
            self.protoinfo = None

        self.mark = cta_mark
        self.id = cta_id
        self.use = cta_use

    def status_name(self):
        s = ''
        for bit, name in IPSBIT_TO_NAME.items():
            if self.status & bit:
                s += '{},'.format(name)
        return s[:-1]

    def __repr__(self):
        s = 'Entry(orig={}, reply={}, status={}'.format(
            self.tuple_orig, self.tuple_reply, self.status_name()
        )
        if self.protoinfo is not None:
            s += ', protoinfo={}'.format(self.protoinfo)
        s += ')'
        return s


class Conntrack(NFCTSocket):
    """
    High level conntrack functions
    """

    def __init__(self, nlm_generator=True, **kwargs):
        super(Conntrack, self).__init__(nlm_generator=nlm_generator, **kwargs)

    def stat(self):
        """Return current statistics per CPU

        Same result than conntrack -S command but a list of dictionaries
        """
        stats = []

        for msg in super(Conntrack, self).stat():
            stats.append({'cpu': msg['res_id']})
            stats[-1].update(
                (k[10:].lower(), v)
                for k, v in msg['attrs']
                if k.startswith('CTA_STATS_')
            )

        return stats

    def count(self):
        """Return current number of conntrack entries

        Same result than /proc/sys/net/netfilter/nf_conntrack_count file
        or conntrack -C command
        """
        for ndmsg in super(Conntrack, self).count():
            return ndmsg.get_attr('CTA_STATS_GLOBAL_ENTRIES')

    def conntrack_max_size(self):
        """
        Return the max size of connection tracking table
        /proc/sys/net/netfilter/nf_conntrack_max
        """
        for ndmsg in super(Conntrack, self).conntrack_max_size():
            return ndmsg.get_attr('CTA_STATS_GLOBAL_MAX_ENTRIES')

    def delete(self, entry):
        if isinstance(entry, ConntrackEntry):
            tuple_orig = entry.tuple_orig
        elif isinstance(entry, NFCTAttrTuple):
            tuple_orig = entry
        else:
            raise NotImplementedError()
        for ndmsg in self.entry('del', tuple_orig=tuple_orig):
            return ndmsg

    def entry(self, cmd, **kwargs):
        for res in super(Conntrack, self).entry(cmd, **kwargs):
            return res

    def dump_entries(
        self,
        mark=None,
        mark_mask=0xFFFFFFFF,
        tuple_orig=None,
        tuple_reply=None,
    ):
        """
        Dump all entries from conntrack table with filters

        Filters can be only part of a conntrack tuple

        :param NFCTAttrTuple tuple_orig: filter on original tuple
        :param NFCTAttrTuple tuple_reply: filter on reply tuple

        Examples::
            # Filter only on tcp connections
            for entry in ct.dump_entries(tuple_orig=NFCTAttrTuple(
                                             proto=socket.IPPROTO_TCP)):
                print("This entry is tcp: {}".format(entry))

            # Filter only on icmp message to 8.8.8.8
            for entry in ct.dump_entries(tuple_orig=NFCTAttrTuple(
                                             proto=socket.IPPROTO_ICMP,
                                             daddr='8.8.8.8')):
                print("This entry is icmp to 8.8.8.8: {}".format(entry))
        """
        for ndmsg in self.dump(
            mark=mark,
            mark_mask=mark_mask,
            tuple_orig=tuple_orig,
            tuple_reply=tuple_reply,
        ):

            if tuple_orig is not None and not tuple_orig.nla_eq(
                ndmsg['nfgen_family'], ndmsg.get_attr('CTA_TUPLE_ORIG')
            ):
                continue

            if tuple_reply is not None and not tuple_reply.nla_eq(
                ndmsg['nfgen_family'], ndmsg.get_attr('CTA_TUPLE_REPLY')
            ):
                continue

            yield ConntrackEntry(
                ndmsg['nfgen_family'],
                ndmsg.get_attr('CTA_TUPLE_ORIG'),
                ndmsg.get_attr('CTA_TUPLE_REPLY'),
                ndmsg.get_attr('CTA_STATUS'),
                ndmsg.get_attr('CTA_TIMEOUT'),
                ndmsg.get_attr('CTA_PROTOINFO'),
                ndmsg.get_attr('CTA_MARK'),
                ndmsg.get_attr('CTA_ID'),
                ndmsg.get_attr('CTA_USE'),
            )



'''
VFS_DQUOT module
================

Usage::

    from pyroute2 import DQuotSocket

    ds = DQuotSocket()
    msgs = ds.get()

Please notice, that `.get()` always returns a list, even if
only one message arrived. To get NLA values::

    msg = msgs[0]
    uid = msg.get_attr('QUOTA_NL_A_EXCESS_ID')
    major = msg.get_attr('QUOTA_NL_A_DEV_MAJOR')
    minor = msg.get_attr('QUOTA_NL_A_DEV_MINOR')
'''
from pr2modules.netlink import genlmsg
from pr2modules.netlink.event import EventSocket
from pr2modules.netlink.nlsocket import Marshal

QUOTA_NL_C_UNSPEC = 0
QUOTA_NL_C_WARNING = 1


class dquotmsg(genlmsg):
    prefix = 'QUOTA_NL_A_'
    nla_map = (
        ('QUOTA_NL_A_UNSPEC', 'none'),
        ('QUOTA_NL_A_QTYPE', 'uint32'),
        ('QUOTA_NL_A_EXCESS_ID', 'uint64'),
        ('QUOTA_NL_A_WARNING', 'uint32'),
        ('QUOTA_NL_A_DEV_MAJOR', 'uint32'),
        ('QUOTA_NL_A_DEV_MINOR', 'uint32'),
        ('QUOTA_NL_A_CAUSED_ID', 'uint64'),
        ('QUOTA_NL_A_PAD', 'uint64'),
    )


class MarshalDQuot(Marshal):
    msg_map = {QUOTA_NL_C_UNSPEC: dquotmsg, QUOTA_NL_C_WARNING: dquotmsg}


class DQuotSocket(EventSocket):
    marshal_class = MarshalDQuot
    genl_family = 'VFS_DQUOT'



from socket import htons

from pr2modules import protocols
from pr2modules.netlink import nla
from pr2modules.netlink.rtnl.tcmsg.common_act import (
    get_tca_action,
    tca_act_prio,
)


def fix_msg(msg, kwarg):
    msg['info'] = htons(
        kwarg.get('protocol', protocols.ETH_P_ALL) & 0xFFFF
    ) | ((kwarg.get('prio', 0) << 16) & 0xFFFF0000)


def get_parameters(kwarg):
    ret = {'attrs': []}
    attrs_map = (
        ('classid', 'TCA_MATCHALL_CLASSID'),
        ('flags', 'TCA_MATCHALL_FLAGS'),
    )

    if kwarg.get('action'):
        ret['attrs'].append(['TCA_MATCHALL_ACT', get_tca_action(kwarg)])

    for k, v in attrs_map:
        r = kwarg.get(k, None)
        if r is not None:
            ret['attrs'].append([v, r])

    return ret


class options(nla):
    nla_map = (
        ('TCA_MATCHALL_UNSPEC', 'none'),
        ('TCA_MATCHALL_CLASSID', 'be32'),
        ('TCA_MATCHALL_ACT', 'tca_act_prio'),
        ('TCA_MATCHALL_FLAGS', 'be32'),
    )

    tca_act_prio = tca_act_prio



#!/usr/bin/env python

# pyroute2 - ss2
# Copyright (C) 2018  Matthias Tafelmeier
#
# ss2 is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.


import argparse
import json
import os
import re
import socket
from socket import AF_INET, AF_UNIX

try:
    import psutil
except ImportError:
    psutil = None
from pr2modules.netlink.diag import (
    SS_ALL,
    SS_CLOSE,
    SS_CLOSE_WAIT,
    SS_CLOSING,
    SS_CONN,
    SS_ESTABLISHED,
    SS_FIN_WAIT1,
    SS_FIN_WAIT2,
    SS_LAST_ACK,
    SS_LISTEN,
    SS_SYN_RECV,
    SS_SYN_SENT,
    SS_TIME_WAIT,
    UDIAG_SHOW_NAME,
    UDIAG_SHOW_PEER,
    UDIAG_SHOW_VFS,
    DiagSocket,
)

try:
    from collections.abc import Callable, Mapping
except ImportError:
    from collections import Callable, Mapping
# UDIAG_SHOW_ICONS,
# UDIAG_SHOW_RQLEN,
# UDIAG_SHOW_MEMINFO


class UserCtxtMap(Mapping):

    _data = {}

    _sk_inode_re = re.compile(r"socket:\[(?P<ino>\d+)\]")

    _proc_sk_fd_cast = "/proc/%d/fd/%d"

    _BUILD_RECURS_PATH = ["inode", "usr", "pid", "fd"]

    def _parse_inode(self, sconn):
        sk_path = self._proc_sk_fd_cast % (sconn.pid, sconn.fd)
        inode = None

        sk_inode_raw = os.readlink(sk_path)
        inode = self._sk_inode_re.search(sk_inode_raw).group('ino')

        if not inode:
            raise RuntimeError("Unexpected kernel sk inode outline")

        return inode

    def __recurs_enter(
        self,
        _sk_inode=None,
        _sk_fd=None,
        _usr=None,
        _pid=None,
        _ctxt=None,
        _recurs_path=[],
    ):

        step = _recurs_path.pop(0)

        if self._BUILD_RECURS_PATH[0] == step:
            if _sk_inode not in self._data.keys():
                self._data[_sk_inode] = {}

        elif self._BUILD_RECURS_PATH[1] == step:
            if _usr not in self._data[_sk_inode].keys():
                self._data[_sk_inode][_usr] = {}

        elif self._BUILD_RECURS_PATH[2] == step:
            if _pid not in self._data[_sk_inode][_usr].keys():
                self._data[_sk_inode][_usr].__setitem__(_pid, _ctxt)

        elif self._BUILD_RECURS_PATH[3] == step:
            self._data[_sk_inode][_usr][_pid]["fds"].append(_sk_fd)

            # end recursion
            return
        else:
            raise RuntimeError("Unexpected step in recursion")

        # descend
        self.__recurs_enter(
            _sk_inode=_sk_inode,
            _sk_fd=_sk_fd,
            _usr=_usr,
            _pid=_pid,
            _ctxt=_ctxt,
            _recurs_path=_recurs_path,
        )

    def _enter_item(self, usr, flow, ctxt):
        if not flow.pid:
            # corner case of eg anonnymous AddressFamily.AF_UNIX
            # sockets
            return

        sk_inode = int(self._parse_inode(flow))
        sk_fd = flow.fd

        recurs_path = list(self._BUILD_RECURS_PATH)

        self.__recurs_enter(
            _sk_inode=sk_inode,
            _sk_fd=sk_fd,
            _usr=usr,
            _pid=flow.pid,
            _ctxt=ctxt,
            _recurs_path=recurs_path,
        )

    def _build(self):
        for flow in psutil.net_connections(kind="all"):
            proc = psutil.Process(flow.pid)
            usr = proc.username()

            ctxt = {"cmd": proc.exe(), "full_cmd": proc.cmdline(), "fds": []}

            self._enter_item(usr, flow, ctxt)

    def __init__(self):
        self._build()

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data.keys())

    def __delitem__(self, key):
        raise RuntimeError("Not implemented")

    def __iter__(self):
        raise RuntimeError("Not implemented")


class Protocol(Callable):
    class Resolver:
        @staticmethod
        def getHost(ip):
            try:
                data = socket.gethostbyaddr(ip)
                host = str(data[0])
                return host
            except Exception:
                # gracefully
                return None

    def __init__(self, sk_states, fmt='json'):
        self._states = sk_states

        fmter = "_fmt_%s" % fmt
        self._fmt = getattr(self, fmter, None)

        def __call__(self, nl_diag_sk, args, usr_ctxt):
            raise RuntimeError('not implemented')

    def _fmt_json(self, refined_stats):
        return json.dumps(refined_stats, indent=4)


class UNIX(Protocol):
    def __init__(self, sk_states=SS_CONN, _fmt='json'):
        super(UNIX, self).__init__(sk_states, fmt=_fmt)

    def __call__(self, nl_diag_sk, args, usr_ctxt):
        sstats = nl_diag_sk.get_sock_stats(
            states=self._states,
            family=AF_UNIX,
            show=(UDIAG_SHOW_NAME | UDIAG_SHOW_VFS | UDIAG_SHOW_PEER),
        )
        refined_stats = self._refine_diag_raw(sstats, usr_ctxt)
        printable = self._fmt(refined_stats)

        print(printable)

    def _refine_diag_raw(self, raw_stats, usr_ctxt):
        refined = {'UNIX': {'flows': []}}

        def vfs_cb(raw_val):
            out = {}
            out['inode'] = raw_val['udiag_vfs_ino']
            out['dev'] = raw_val['udiag_vfs_dev']

            return out

        k_idx = 0
        val_idx = 1
        cb_idx = 1

        idiag_attr_refine_map = {
            'UNIX_DIAG_NAME': ('path_name', None),
            'UNIX_DIAG_VFS': ('vfs', vfs_cb),
            'UNIX_DIAG_PEER': ('peer_inode', None),
            'UNIX_DIAG_SHUTDOWN': ('shutdown', None),
        }

        for raw_flow in raw_stats:
            vessel = {}
            vessel['inode'] = raw_flow['udiag_ino']

            for attr in raw_flow['attrs']:
                attr_k = attr[k_idx]
                attr_val = attr[val_idx]
                k = idiag_attr_refine_map[attr_k][k_idx]
                cb = idiag_attr_refine_map[attr_k][cb_idx]

                if cb:
                    attr_val = cb(attr_val)

                vessel[k] = attr_val

            refined['UNIX']['flows'].append(vessel)

        if usr_ctxt:
            for flow in refined['UNIX']['flows']:
                try:
                    sk_inode = flow['inode']
                    flow['usr_ctxt'] = usr_ctxt[sk_inode]
                except KeyError:
                    # might define sentinel val
                    pass

        return refined


class TCP(Protocol):

    INET_DIAG_MEMINFO = 1
    INET_DIAG_INFO = 2
    INET_DIAG_VEGASINFO = 3
    INET_DIAG_CONG = 4

    def __init__(self, sk_states=SS_CONN, _fmt='json'):
        super(TCP, self).__init__(sk_states, fmt=_fmt)

        IDIAG_EXT_FLAGS = [
            self.INET_DIAG_MEMINFO,
            self.INET_DIAG_INFO,
            self.INET_DIAG_VEGASINFO,
            self.INET_DIAG_CONG,
        ]

        self.ext_f = 0
        for f in IDIAG_EXT_FLAGS:
            self.ext_f |= 1 << (f - 1)

    def __call__(self, nl_diag_sk, args, usr_ctxt):
        sstats = nl_diag_sk.get_sock_stats(
            states=self._states, family=AF_INET, extensions=self.ext_f
        )
        refined_stats = self._refine_diag_raw(sstats, args.resolve, usr_ctxt)
        printable = self._fmt(refined_stats)

        print(printable)

    def _refine_diag_raw(self, raw_stats, do_resolve, usr_ctxt):

        refined = {'TCP': {'flows': []}}

        idiag_refine_map = {
            'src': 'idiag_src',
            'dst': 'idiag_dst',
            'src_port': 'idiag_sport',
            'dst_port': 'idiag_dport',
            'inode': 'idiag_inode',
            'iface_idx': 'idiag_if',
            'retrans': 'idiag_retrans',
        }

        for raw_flow in raw_stats:
            vessel = {}
            for k1, k2 in idiag_refine_map.items():
                vessel[k1] = raw_flow[k2]

            for ext_bundle in raw_flow['attrs']:
                vessel = self._refine_extension(vessel, ext_bundle)

            refined['TCP']['flows'].append(vessel)

        if usr_ctxt:
            for flow in refined['TCP']['flows']:
                try:
                    sk_inode = flow['inode']
                    flow['usr_ctxt'] = usr_ctxt[sk_inode]
                except KeyError:
                    # might define sentinel val
                    pass

        if do_resolve:
            for flow in refined['TCP']['flows']:
                src_host = Protocol.Resolver.getHost(flow['src'])
                if src_host:
                    flow['src_host'] = src_host

                dst_host = Protocol.Resolver.getHost(flow['dst'])
                if dst_host:
                    flow['dst_host'] = dst_host

        return refined

    def _refine_extension(self, vessel, raw_ext):
        k, content = raw_ext
        ext_refine_map = {
            'meminfo': {
                'r': 'idiag_rmem',
                'w': 'idiag_wmem',
                'f': 'idiag_fmem',
                't': 'idiag_tmem',
            }
        }

        if k == 'INET_DIAG_MEMINFO':
            mem_k = 'meminfo'
            vessel[mem_k] = {}
            for k1, k2 in ext_refine_map[mem_k].items():
                vessel[mem_k][k1] = content[k2]

        elif k == 'INET_DIAG_CONG':
            vessel['cong_algo'] = content

        elif k == 'INET_DIAG_INFO':
            vessel = self._refine_tcp_info(vessel, content)

        elif k == 'INET_DIAG_SHUTDOWN':
            pass

        return vessel

    # interim approach
    # tcpinfo call backs
    class InfoCbCore:

        # normalizer
        @staticmethod
        def rto_n_cb(key, value, **ctx):
            out = None
            if value != 3000000:
                out = value / 1000.0

            return out

        @staticmethod
        def generic_1k_n_cb(key, value, **ctx):
            return value / 1000.0

        # predicates
        @staticmethod
        def snd_thresh_p_cb(key, value, **ctx):
            if value < 0xFFFF:
                return value

            return None

        @staticmethod
        def rtt_p_cb(key, value, **ctx):
            tcp_info_raw = ctx['raw']

            try:
                if (
                    tcp_info_raw['tcpv_enabled'] != 0
                    and tcp_info_raw['tcpv_rtt'] != 0x7FFFFFFF
                ):
                    return tcp_info_raw['tcpv_rtt']
            except KeyError:
                # ill practice, yet except quicker path
                pass

            return tcp_info_raw['tcpi_rtt'] / 1000.0

        # converter
        @staticmethod
        def state_c_cb(key, value, **ctx):
            state_str_map = {
                SS_ESTABLISHED: "established",
                SS_SYN_SENT: "syn-sent",
                SS_SYN_RECV: "syn-recv",
                SS_FIN_WAIT1: "fin-wait-1",
                SS_FIN_WAIT2: "fin-wait-2",
                SS_TIME_WAIT: "time-wait",
                SS_CLOSE: "unconnected",
                SS_CLOSE_WAIT: "close-wait",
                SS_LAST_ACK: "last-ack",
                SS_LISTEN: "listening",
                SS_CLOSING: "closing",
            }

            return state_str_map[value]

        @staticmethod
        def opts_c_cb(key, value, **ctx):
            tcp_info_raw = ctx['raw']

            # tcp_info opt flags
            TCPI_OPT_TIMESTAMPS = 1
            TCPI_OPT_SACK = 2
            TCPI_OPT_ECN = 8

            out = []

            opts = tcp_info_raw['tcpi_options']
            if opts & TCPI_OPT_TIMESTAMPS:
                out.append("ts")
            if opts & TCPI_OPT_SACK:
                out.append("sack")
            if opts & TCPI_OPT_ECN:
                out.append("ecn")

            return out

    def _refine_tcp_info(self, vessel, tcp_info_raw):

        ti = TCP.InfoCbCore

        info_refine_tabl = {
            'tcpi_state': ('state', ti.state_c_cb),
            'tcpi_pmtu': ('pmtu', None),
            'tcpi_retrans': ('retrans', None),
            'tcpi_ato': ('ato', ti.generic_1k_n_cb),
            'tcpi_rto': ('rto', ti.rto_n_cb),
            # TODO consider wscale baking
            'tcpi_snd_wscale': ('snd_wscale', None),
            'tcpi_rcv_wscale': ('rcv_wscale', None),
            # TODO bps baking
            'tcpi_snd_mss': ('snd_mss', None),
            'tcpi_snd_cwnd': ('snd_cwnd', None),
            'tcpi_snd_ssthresh': ('snd_ssthresh', ti.snd_thresh_p_cb),
            # TODO consider rtt agglomeration - needs nesting
            'tcpi_rtt': ('rtt', ti.rtt_p_cb),
            'tcpi_rttvar': ('rttvar', ti.generic_1k_n_cb),
            'tcpi_rcv_rtt': ('rcv_rtt', ti.generic_1k_n_cb),
            'tcpi_rcv_space': ('rcv_space', None),
            'tcpi_options': ('opts', ti.opts_c_cb),
            # unclear, NB not in use by iproute2 ss latest
            'tcpi_last_data_sent': ('last_data_sent', None),
            'tcpi_rcv_ssthresh': ('rcv_ssthresh', None),
            'tcpi_rcv_ssthresh': ('rcv_ssthresh', None),
            'tcpi_segs_in': ('segs_in', None),
            'tcpi_segs_out': ('segs_out', None),
            'tcpi_data_segs_in': ('data_segs_in', None),
            'tcpi_data_segs_out': ('data_segs_out', None),
            'tcpi_lost': ('lost', None),
            'tcpi_notsent_bytes': ('notsent_bytes', None),
            'tcpi_rcv_mss': ('rcv_mss', None),
            'tcpi_pacing_rate': ('pacing_rate', None),
            'tcpi_retransmits': ('retransmits', None),
            'tcpi_min_rtt': ('min_rtt', None),
            'tcpi_rwnd_limited': ('rwnd_limited', None),
            'tcpi_max_pacing_rate': ('max_pacing_rate', None),
            'tcpi_probes': ('probes', None),
            'tcpi_reordering': ('reordering', None),
            'tcpi_last_data_recv': ('last_data_recv', None),
            'tcpi_bytes_received': ('bytes_received', None),
            'tcpi_fackets': ('fackets', None),
            'tcpi_last_ack_recv': ('last_ack_recv', None),
            'tcpi_last_ack_sent': ('last_ack_sent', None),
            'tcpi_unacked': ('unacked', None),
            'tcpi_sacked': ('sacked', None),
            'tcpi_bytes_acked': ('bytes_acked', None),
            'tcpi_delivery_rate_app_limited': (
                'delivery_rate_app_limited',
                None,
            ),
            'tcpi_delivery_rate': ('delivery_rate', None),
            'tcpi_sndbuf_limited': ('sndbuf_limited', None),
            'tcpi_ca_state': ('ca_state', None),
            'tcpi_busy_time': ('busy_time', None),
            'tcpi_total_retrans': ('total_retrans', None),
            'tcpi_advmss': ('advmss', None),
            'tcpi_backoff': (None, None),
            'tcpv_enabled': (None, 'skip'),
            'tcpv_rttcnt': (None, 'skip'),
            'tcpv_rtt': (None, 'skip'),
            'tcpv_minrtt': (None, 'skip'),
            # BBR
            'bbr_bw_lo': ('bbr_bw_lo', None),
            'bbr_bw_hi': ('bbr_bw_hi', None),
            'bbr_min_rtt': ('bbr_min_rtt', None),
            'bbr_pacing_gain': ('bbr_pacing_gain', None),
            'bbr_cwnd_gain': ('bbr_cwnd_gain', None),
            # DCTCP
            'dctcp_enabled': ('dctcp_enabled', None),
            'dctcp_ce_state': ('dctcp_ce_state', None),
            'dctcp_alpha': ('dctcp_alpha', None),
            'dctcp_ab_ecn': ('dctcp_ab_ecn', None),
            'dctcp_ab_tot': ('dctcp_ab_tot', None),
        }
        k_idx = 0
        cb_idx = 1

        info_k = 'tcp_info'
        vessel[info_k] = {}

        # BUG - pyroute2 diag - seems always last info instance from kernel
        if type(tcp_info_raw) != str:
            for k, v in tcp_info_raw.items():
                if k not in info_refine_tabl:
                    continue
                refined_k = info_refine_tabl[k][k_idx]
                cb = info_refine_tabl[k][cb_idx]
                refined_v = v
                if cb and cb == 'skip':
                    continue
                elif cb:
                    ctx = {'raw': tcp_info_raw}
                    refined_v = cb(k, v, **ctx)

                vessel[info_k][refined_k] = refined_v

        return vessel


def prepare_args():
    parser = argparse.ArgumentParser(
        description="""
                                     ss2 - socket statistics depictor meant as
                                     a complete and convenient surrogate for
                                     iproute2/misc/ss2"""
    )
    parser.add_argument(
        '-x',
        '--unix',
        help='Display Unix domain sockets.',
        action='store_true',
    )
    parser.add_argument(
        '-t', '--tcp', help='Display TCP sockets.', action='store_true'
    )
    parser.add_argument(
        '-l',
        '--listen',
        help='Display listening sockets.',
        action='store_true',
    )
    parser.add_argument(
        '-a', '--all', help='Display all sockets.', action='store_true'
    )
    parser.add_argument(
        '-p',
        '--process',
        help='show socket holding context',
        action='store_true',
    )
    parser.add_argument(
        '-r',
        '--resolve',
        help='resolve host names in addition',
        action='store_true',
    )

    args = parser.parse_args()

    return args


def run(args=None):

    if psutil is None:
        raise RuntimeError('ss2 requires python-psutil >= 5.0 to run')

    if not args:
        args = prepare_args()

    _states = SS_CONN
    if args.listen:
        _states = 1 << SS_LISTEN
    if args.all:
        _states = SS_ALL

    protocols = []
    if args.tcp:
        protocols.append(TCP(sk_states=_states))

    if args.unix:
        protocols.append(UNIX(sk_states=_states))

    if not protocols:
        raise RuntimeError('not implemented - ss2 in fledging mode')

    _user_ctxt_map = None
    if args.process:
        _user_ctxt_map = UserCtxtMap()

    with DiagSocket() as ds:
        ds.bind()
        for p in protocols:
            p(ds, args, _user_ctxt_map)


if __name__ == "__main__":
    run()



#
# Author: Angus Lees <gus@inodes.org>
#
# Backported from a Neutron privsep proposal with the
# permission of the author.
#

from __future__ import absolute_import

import functools
import socket
import types

try:
    import cPickle as pickle
except ImportError:
    import pickle

from pr2modules import config as config

_socketmethods = (
    'bind',
    'close',
    'connect',
    'connect_ex',
    'listen',
    'getpeername',
    'getsockname',
    'getsockopt',
    'makefile',
    'recv',
    'recvfrom',
    'recv_into',
    'recvfrom_into',
    'send',
    'sendto',
    'sendall',
    'setsockopt',
    'setblocking',
    'settimeout',
    'gettimeout',
    'shutdown',
)


def _forward(name, self, *args, **kwargs):
    return getattr(self._sock, name)(*args, **kwargs)


class _SocketWrapper(object):
    """eventlet-monkeypatch friendly socket class"""

    def __init__(self, *args, **kwargs):
        _sock = kwargs.get('_sock', None) or socket.socket(*args, **kwargs)
        self._sock = _sock
        for name in _socketmethods:
            f = functools.partial(_forward, name)
            f.__name__ = name
            setattr(_SocketWrapper, name, types.MethodType(f, self))

    def fileno(self):
        return self._sock.fileno()

    def dup(self):
        return self.__class__(_sock=self._sock.dup())


class _MpConnection(object):
    """Highly limited multiprocessing.Connection alternative"""

    def __init__(self, sock):
        sock.setblocking(True)
        self.sock = sock

    def fileno(self):
        return self.sock.fileno()

    def send(self, obj):
        pickle.dump(obj, self, protocol=-1)

    def write(self, s):
        self.sock.sendall(s)

    def recv(self):
        return pickle.load(self)

    def read(self, n):
        return self.sock.recv(n)

    def readline(self):
        buf = b''
        c = None
        while c != b'\n':
            c = self.sock.recv(1)
            buf += c
        return buf

    def close(self):
        self.sock.close()


def _MultiprocessingPipe():
    """multiprocess.Pipe reimplementation that uses MpConnection wrapper"""

    s1, s2 = socket.socketpair()
    return (_MpConnection(s1), _MpConnection(s2))


def asyncio_config():
    config.SocketBase = _SocketWrapper
    config.MpPipe = _MultiprocessingPipe
    config.ipdb_nl_async = False



from pr2modules.netlink import rtnl
from pr2modules.netlink.nlsocket import Marshal
from pr2modules.netlink.rtnl.fibmsg import fibmsg
from pr2modules.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pr2modules.netlink.rtnl.ifinfmsg import ifinfmsg
from pr2modules.netlink.rtnl.ifstatsmsg import ifstatsmsg
from pr2modules.netlink.rtnl.ndmsg import ndmsg
from pr2modules.netlink.rtnl.ndtmsg import ndtmsg
from pr2modules.netlink.rtnl.nsidmsg import nsidmsg
from pr2modules.netlink.rtnl.rtmsg import rtmsg
from pr2modules.netlink.rtnl.tcmsg import tcmsg


class MarshalRtnl(Marshal):
    msg_map = {
        rtnl.RTM_NEWLINK: ifinfmsg,
        rtnl.RTM_DELLINK: ifinfmsg,
        rtnl.RTM_GETLINK: ifinfmsg,
        rtnl.RTM_SETLINK: ifinfmsg,
        rtnl.RTM_NEWADDR: ifaddrmsg,
        rtnl.RTM_DELADDR: ifaddrmsg,
        rtnl.RTM_GETADDR: ifaddrmsg,
        rtnl.RTM_NEWROUTE: rtmsg,
        rtnl.RTM_DELROUTE: rtmsg,
        rtnl.RTM_GETROUTE: rtmsg,
        rtnl.RTM_NEWRULE: fibmsg,
        rtnl.RTM_DELRULE: fibmsg,
        rtnl.RTM_GETRULE: fibmsg,
        rtnl.RTM_NEWNEIGH: ndmsg,
        rtnl.RTM_DELNEIGH: ndmsg,
        rtnl.RTM_GETNEIGH: ndmsg,
        rtnl.RTM_NEWQDISC: tcmsg,
        rtnl.RTM_DELQDISC: tcmsg,
        rtnl.RTM_GETQDISC: tcmsg,
        rtnl.RTM_NEWTCLASS: tcmsg,
        rtnl.RTM_DELTCLASS: tcmsg,
        rtnl.RTM_GETTCLASS: tcmsg,
        rtnl.RTM_NEWTFILTER: tcmsg,
        rtnl.RTM_DELTFILTER: tcmsg,
        rtnl.RTM_GETTFILTER: tcmsg,
        rtnl.RTM_NEWNEIGHTBL: ndtmsg,
        rtnl.RTM_GETNEIGHTBL: ndtmsg,
        rtnl.RTM_SETNEIGHTBL: ndtmsg,
        rtnl.RTM_NEWNSID: nsidmsg,
        rtnl.RTM_DELNSID: nsidmsg,
        rtnl.RTM_GETNSID: nsidmsg,
        rtnl.RTM_NEWSTATS: ifstatsmsg,
        rtnl.RTM_GETSTATS: ifstatsmsg,
        rtnl.RTM_NEWLINKPROP: ifinfmsg,
        rtnl.RTM_DELLINKPROP: ifinfmsg,
    }

    def fix_message(self, msg):
        # FIXME: pls do something with it
        try:
            msg['event'] = rtnl.RTM_VALUES[msg['header']['type']]
        except:
            pass



from pr2modules.netlink import nla
from pr2modules.netlink.rtnl import TC_H_ROOT

parent = TC_H_ROOT


class options(nla):
    fields = (('limit', 'i'),)


def get_parameters(kwarg):
    return kwarg



import errno
import os
import struct
from fcntl import ioctl

from pr2modules import config
from pr2modules.netlink.exceptions import NetlinkError
from pr2modules.netlink.rtnl.ifinfmsg import (
    IFT_MULTI_QUEUE,
    IFT_NO_PI,
    IFT_ONE_QUEUE,
    IFT_TAP,
    IFT_TUN,
    IFT_VNET_HDR,
    RTM_NEWLINK,
)
from pr2modules.netlink.rtnl.ifinfmsg.sync import sync

IFNAMSIZ = 16

TUNDEV = '/dev/net/tun'
PLATFORMS = (
    'i386',
    'i686',
    'x86_64',
    'armv6l',
    'armv7l',
    's390x',
    'aarch64',
    'loongarch64',
)
if config.machine in PLATFORMS:
    TUNSETIFF = 0x400454CA
    TUNSETPERSIST = 0x400454CB
    TUNSETOWNER = 0x400454CC
    TUNSETGROUP = 0x400454CE
elif config.machine in ('ppc64', 'mips'):
    TUNSETIFF = 0x800454CA
    TUNSETPERSIST = 0x800454CB
    TUNSETOWNER = 0x800454CC
    TUNSETGROUP = 0x800454CE
else:
    TUNSETIFF = None


@sync
def manage_tun(msg):

    if TUNSETIFF is None:
        raise NetlinkError(errno.EOPNOTSUPP, 'Arch not supported')

    if msg['header']['type'] != RTM_NEWLINK:
        raise NetlinkError(errno.EOPNOTSUPP, 'Unsupported event')

    ifru_flags = 0
    linkinfo = msg.get_attr('IFLA_LINKINFO')
    infodata = linkinfo.get_attr('IFLA_INFO_DATA')

    if infodata.get_attr('IFLA_TUN_TYPE') == 1:
        ifru_flags |= IFT_TUN
    elif infodata.get_attr('IFLA_TUN_TYPE') == 2:
        ifru_flags |= IFT_TAP
    else:
        raise ValueError('invalid mode')
    if not infodata.get_attr('IFLA_TUN_PI'):
        ifru_flags |= IFT_NO_PI
    if infodata.get_attr('IFLA_TUN_VNET_HDR'):
        ifru_flags |= IFT_VNET_HDR
    if infodata.get_attr('IFLA_TUN_MULTI_QUEUE'):
        ifru_flags |= IFT_MULTI_QUEUE

    ifr = msg.get_attr('IFLA_IFNAME')
    if len(ifr) > IFNAMSIZ:
        raise ValueError('ifname too long')
    ifr += (IFNAMSIZ - len(ifr)) * '\0'
    ifr = ifr.encode('ascii')
    ifr += struct.pack('H', ifru_flags)

    user = infodata.get_attr('IFLA_TUN_OWNER')
    group = infodata.get_attr('IFLA_TUN_GROUP')
    #
    fd = os.open(TUNDEV, os.O_RDWR)
    try:
        ioctl(fd, TUNSETIFF, ifr)
        if user is not None:
            ioctl(fd, TUNSETOWNER, user)
        if group is not None:
            ioctl(fd, TUNSETGROUP, group)
        ioctl(fd, TUNSETPERSIST, 1)
    except Exception:
        raise
    finally:
        os.close(fd)


@sync
def manage_tuntap(msg):

    if TUNSETIFF is None:
        raise NetlinkError(errno.EOPNOTSUPP, 'Arch not supported')

    if msg['header']['type'] != RTM_NEWLINK:
        raise NetlinkError(errno.EOPNOTSUPP, 'Unsupported event')

    ifru_flags = 0
    linkinfo = msg.get_attr('IFLA_LINKINFO')
    infodata = linkinfo.get_attr('IFLA_INFO_DATA')

    flags = infodata.get_attr('IFTUN_IFR', None)
    if infodata.get_attr('IFTUN_MODE') == 'tun':
        ifru_flags |= IFT_TUN
    elif infodata.get_attr('IFTUN_MODE') == 'tap':
        ifru_flags |= IFT_TAP
    else:
        raise ValueError('invalid mode')
    if flags is not None:
        if flags['no_pi']:
            ifru_flags |= IFT_NO_PI
        if flags['one_queue']:
            ifru_flags |= IFT_ONE_QUEUE
        if flags['vnet_hdr']:
            ifru_flags |= IFT_VNET_HDR
        if flags['multi_queue']:
            ifru_flags |= IFT_MULTI_QUEUE
    ifr = msg.get_attr('IFLA_IFNAME')
    if len(ifr) > IFNAMSIZ:
        raise ValueError('ifname too long')
    ifr += (IFNAMSIZ - len(ifr)) * '\0'
    ifr = ifr.encode('ascii')
    ifr += struct.pack('H', ifru_flags)

    user = infodata.get_attr('IFTUN_UID')
    group = infodata.get_attr('IFTUN_GID')
    #
    fd = os.open(TUNDEV, os.O_RDWR)
    try:
        ioctl(fd, TUNSETIFF, ifr)
        if user is not None:
            ioctl(fd, TUNSETOWNER, user)
        if group is not None:
            ioctl(fd, TUNSETGROUP, group)
        ioctl(fd, TUNSETPERSIST, 1)
    except Exception:
        raise
    finally:
        os.close(fd)
